/*
 *
 * apps/videoServer/vsConfiguration.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "vsConfiguration.H"

#include <nucleo/utils/StringUtils.H>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <string>
#include <iostream>

using namespace nucleo ;

static std::string
getConfigValue(const std::string &key, const std::string &config) {
  std::string::size_type p = 0 ;
  for (;;) {
    p = config.find(key, p) ;
    if (p==std::string::npos) return "" ;
    if (p==0 || config[p-1]=='\n' || config[p-1]=='\r') break ;
    p++ ;
  }

  std::string::size_type q = config.find_first_of("\n\r",p) ;
  std::string value ;
  if (q!=std::string::npos) value.assign(config,p,q-p) ;
  else value.assign(config,p,config.length()-p) ;
  std::string tmp = extractNextWord(value) ;
  return value ;
}

namespace videoServer {

  vsConfiguration::vsConfiguration(const char *filename) {

    // Default values

    debug_level = 10 ;
    char buffer[256] ;
    getcwd(buffer, 256) ;
    root = buffer ;
    repository = buffer ;
    tcp_port = 5555 ;
    tcp_backlog = 3 ;
    notifier = "notifier" ;
    video_device = "anydev" ;
    video_node = "anynode" ;
    video_locked = 0 ;
    video_size = "" ;
    video_quality = 60 ;
    video_pause = 0.03 ;
    video_length = 5000 ;
    video_options = "" ;

    // Try to read the configuration from the file

    struct stat statinfo ;
    if (stat(filename, &statinfo)==-1) {
	 std::cerr << "Unable to read configuration from " << filename << std::endl ;
	 return ;
    }

    unsigned int size = statinfo.st_size ;
    char *ptr = new char [size+1] ;
    int fd = open(filename, O_RDONLY) ;
    read(fd, ptr, size) ;
    ptr[size] = '\0' ;
    close(fd) ;
    std::string configuration = ptr ;
    delete [] ptr ;

    // Get the parameters from the configuration

    std::string param = getConfigValue("DebugLevel", configuration) ;
    if (param!="") debug_level = atoi(param.c_str()) ;

    param = getConfigValue("ServerRoot", configuration) ;
    if (param!="") {
	 std::string::size_type last = param.length()-1 ;
	 if (param[last]=='/') param.erase(last, 1) ;
	 root = param ;  
    }

    param = getConfigValue("MoviesRepository", configuration) ;
    if (param!="") {
	 std::string::size_type last = param.length()-1 ;
	 if (param[last]=='/') param.erase(last, 1) ;
	 repository = param ;  
    }

    param = getConfigValue("TcpPort", configuration) ;
    if (param!="") tcp_port = atoi(param.c_str()) ;

    param = getConfigValue("TcpBackLog", configuration) ;
    if (param!="") tcp_backlog = atoi(param.c_str()) ;

    param = getConfigValue("Notifier", configuration) ;
    if (param!="") notifier = param ;

    param = getConfigValue("DefaultDevice", configuration) ;
    if (param!="") video_device = param ;

    param = getConfigValue("DefaultNode", configuration) ;
    if (param!="") video_node = param ;

    param = getConfigValue("DefaultSize", configuration) ;
    if (param!="") video_size = param ;

    param = getConfigValue("DefaultLockState", configuration) ;
    if (param!="") video_locked = atoi(param.c_str()) ;

    param = getConfigValue("DefaultQuality", configuration) ;
    if (param!="") video_quality = atoi(param.c_str()) ;

    param = getConfigValue("DefaultPause", configuration) ;
    if (param!="") video_pause = atof(param.c_str()) ;

    param = getConfigValue("DefaultVideoLength", configuration) ;
    if (param!="") video_length = atoi(param.c_str()) ;

    param = getConfigValue("VideoInOptions", configuration) ;
    if (param!="") video_options = param ;

    if (debug_level>=1) {
	 std::cerr << "DebugLevel         = " << debug_level << std::endl ;
	 std::cerr << "ServerRoot         = \"" << root << "\"" << std::endl ;
	 std::cerr << "MoviesRepository   = \"" << repository << "\"" << std::endl ;
	 std::cerr << "TcpPort            = " << tcp_port << std::endl ;
	 std::cerr << "TcpBackLog         = " << tcp_backlog << std::endl ;
	 std::cerr << "Notifier           = \"" << notifier << "\"" << std::endl ;
	 std::cerr << "DefaultDevice      = \"" << video_device << "\"" << std::endl ;
	 std::cerr << "DefaultNode        = \"" << video_node << "\"" << std::endl ;
	 std::cerr << "DefaultSize        = " << video_size << std::endl ;
	 std::cerr << "DefaultLockState   = " << video_locked << std::endl ;
	 std::cerr << "DefaultQuality     = " << video_quality << std::endl ;
	 std::cerr << "DefaultPause       = " << video_pause << std::endl ;
	 std::cerr << "DefaultVideoLength = " << video_length << std::endl ;
	 std::cerr << "VideoInOptions     = " << video_options << std::endl ;
    }

    chdir(root.c_str()) ;
  }

}
