/*
 *
 * videoServer/vsPostNotifier.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "vsConfiguration.H"

#include <nucleo/core/URL.H>
#include <nucleo/core/PluginManager.H>
#include <nucleo/core/FileKeeper.H>
#include <nucleo/core/ReactiveEngine.H>

#include <nucleo/network/http/HttpMessage.H>
#include <nucleo/network/tcp/TcpConnection.H>

#include <nucleo/image/source/ImageSource.H>
#include <nucleo/image/sink/ImageSink.H>

#include <nucleo/image/sink/serverpushImageSink.H>
#include <nucleo/image/sink/nudpImageSink.H>

#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

using namespace nucleo ;
using namespace videoServer ;

// -----------------------------------------------------------------------------------

static void
sendImage(const vsConfiguration &config, int fd, std::string &args) {
  ImageSource *source = ImageSource::create(args, Image::JPEG) ;
  Image img ;
  source->start() ;
  source->waitForImage(&img) ;
    
  std::string reply = "HTTP/1.0 200 Ok" ;
  reply.append(oneCRLF) ;
  reply.append("Expires: Thu, 01 Dec 1994 16:00:00 GMT") ;
  reply.append(oneCRLF) ;
  reply.append("Content-Type: image/jpeg") ;
  reply.append(twoCRLF) ;
  if (write(fd, (char*) reply.c_str(), reply.length()) != -1 ) {
    // std::cerr << img.getWidth() << " " << img.getHeight() << " " << img.getSize() << std::endl ;
    write(fd, (char*) img.getData(), img.getSize()) ;
  } else if (config.debug_level>=1)
    std::cerr << "sendImage failed" << std::endl ;

  delete source ;
} 

// -----------------------------------------------------------------------------------

typedef struct {
  char *extension ;
  char *mimetype ;
} MimeType ;

static MimeType MimeTypes[] = {
  {".gif","image/gif"},
  {".jpg","image/jpeg"},
  {".html","text/html"},
  {".ps","application/postscript"},
  {".pdf","application/pdf"},
  {".mov","video/quicktime"},
  {".mpeg","video/mpeg"},
  {".mpg","video/mpeg"},
  {".txt","text/plain"},
  {".vss","video/x-nucleo"},
  {0,"application/octet-stream"}
} ;

static void
sendFile(const vsConfiguration &config, int fd, std::string &path) {
  struct stat statinfo ;
  char *f = (char *)path.c_str() ;

  if (stat(f, &statinfo)!=-1) {
    // Get the data
    unsigned int size = statinfo.st_size ;
    unsigned char *ptr = new unsigned char [size] ;
    int datafd = open(f, O_RDONLY) ;
    read(datafd, (char*) ptr, size) ;
    close(datafd) ;

    // Find the right content-type
    char *ext=0, *mtyp="text/plain" ;
    int i ;
    for (i=strlen(f); i>=0 && f[i]!='.'; i--) ;
    if (i>-1) ext = f+i ;

    if (ext) {
	 for (i=0; MimeTypes[i].extension!=0 && strcmp(ext,MimeTypes[i].extension); i++) ;
	 mtyp = MimeTypes[i].mimetype ;
    }

    // Send the data
    char reply[256] ;
    sprintf(reply,"HTTP/1.0 200 OK\r\nExpires: Thu, 01 Dec 1994 16:00:00 GMT\r\nContent-type: %s\r\nContent-Length: %d\r\n\r\n",mtyp,size) ;
    if ( write(fd, reply, strlen(reply)) != -1) write(fd, ptr, size) ;
    delete [] ptr ;
  } else {
    char reply[256] ;
    sprintf(reply,"HTTP/1.0 404 Not Found\r\nExpires: Thu, 01 Dec 1994 16:00:00 GMT\r\n\r\n") ;
    write(fd, reply, strlen(reply)) ;
    if (config.debug_level>=1) std::cerr << "sendFile failed (" << path << ")" << std::endl ;
  }
}

// -----------------------------------------------------------------------------------

static void
redirect(const vsConfiguration &config, int fd, std::string &args) {
  std::string reply = "HTTP/1.0 302 Moved Temporarily" ;
  reply.append(oneCRLF) ;
  reply.append("Location: ") ;
  reply.append(args) ;
  reply.append(twoCRLF) ;
  if (write(fd, (char*) reply.c_str(), reply.length()) == -1 )
    if (config.debug_level>=1)
	 std::cerr << "redirect failed" << std::endl ;
}

// -----------------------------------------------------------------------------------

namespace videoServer {

  void vsPostNotifier(const vsConfiguration &config,
				  const std::string &service,
				  TcpConnection *client) {
    std::string args = service ;
    std::string method = extractNextWord(args) ;

    int fd = client->getFd() ;

    if (method=="push" || method=="vstp") {
	 ImageSource *source = 0 ;
	 ImageSink *sink = 0 ;
	
	 FileKeeper *signal = FileKeeper::create(fd, FileKeeper::R) ;

	 if (method=="vstp") {
	   std::string sourceUrl = extractNextWord(args) ;
	   std::string hostOrGroup = extractNextWord(args) ;
	   std::string port = extractNextWord(args) ;
	   source = ImageSource::create(sourceUrl, Image::JPEG) ;
	   sink = (ImageSink *) new nudpImageSink(hostOrGroup.c_str(), atoi(port.c_str()),
									  Image::JPEG, 80, 1) ;
	 } else {
	   source = ImageSource::create(args, Image::JPEG) ;
	   // std::cerr << "source = " << source << std::endl ;
	   sink = (ImageSink *) new serverpushImageSink(fd) ;
	 }

	 Image img ;
	 int nbImages=0 ;

	 for (source->start(), sink->start();
		 source->isActive() && sink->isActive();) {
	   ReactiveEngine::step() ;
	   if (source->getNextImage(&img)) {
		sink->handle(&img) ;
		++nbImages ;
	   }
	   if (signal->getState()==FileKeeper::R) break ;
	 } 

	 delete source ;
	 delete sink ;

	 if (method=="vstp") {
	   std::stringstream msg ;
	   msg << "HTTP/1.0 200 OK\r\n" ;
	   msg << "Expires: Thu, 01 Dec 1994 16:00:00 GMT\r\n" ;
	   msg << "Content-type: text/plain\r\n\r\n" ;
	   msg << nbImages << " images sent" << std::endl ;
	   std::string s = msg.str() ;
	   write(fd, s.c_str(), s.size()) ;
	 }

	 if (config.debug_level>=1) std::cerr << nbImages << " images sent" << std::endl ;

	 return ;
    }

    if (method=="grab")
	 sendImage(config, fd, args) ;
    else if (method=="sendFile")
	 sendFile(config, fd, args) ;
    else if (method=="redirect")
	 redirect(config, fd, args) ;
  }

}
