/*
 *
 * videoServer/vsPreNotifier.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "vsConfiguration.H"

#include <nucleo/core/URL.H>
#include <nucleo/utils/StringUtils.H>
#include <nucleo/utils/FileUtils.H>

#include <string>
#include <sstream>

using namespace nucleo ;
using namespace videoServer ;

static void
decodeVideoInURL(const URL& url,
			  std::string *device, std::string *node,
			  std::string *size,
			  int *length, float *pause,
			  int *locked,
			  int *quality) {
  std::string path = url.getPath() ;

  if (path!="") {
    std::string::size_type n=0 ;
    for (std::string::size_type i=1; i<path.length(); ++i) {
	 if (path[i]=='/') n=i+1 ;
    }
    device->assign(path,1,n-2) ;
    if (n) node->assign(path,n,path.length()-n+1) ;
  }

  std::string query = url.getQuery() ;

  std::string param ;
  if (URL::getQueryArg(query, "size", &param)) *size = param ;
  if (URL::getQueryArg(query, "length", &param)) *length = atoi(param.c_str()) ;
  if (URL::getQueryArg(query, "pause", &param)) *pause = atof(param.c_str()) ;
  if (URL::getQueryArg(query, "locked", &param)) *locked = atoi(param.c_str()) ;
  if (URL::getQueryArg(query, "quality", &param)) *quality = atoi(param.c_str()) ;
}

// -------------------------------------------------------------------------------

static std::string
parseSource(const vsConfiguration &config, const std::string &path, const std::string &query) {
  std::stringstream source ;
  std::string param ;

  if ( path.substr(0,6)=="/photo" || path.substr(0,6)=="/video" ) {

    // Local source

    source << "videoin:/" ;
    source << (URL::getQueryArg(query, "device", &param) ? param : config.video_device) ;

    source << "/" ;
    source << (URL::getQueryArg(query, "node", &param) ? param : config.video_node) ;

    source << "?" ;
    if (config.video_options!="") source << config.video_options << "&" ;

    source << "size=" ;
    source << (URL::getQueryArg(query, "size", &param) ? param : config.video_size) ;

    source << "&quality=" ;
    if (URL::getQueryArg(query, "quality", &param)) source << param ; else source << config.video_quality ;

    source << "&locked=" ;
    if (URL::getQueryArg(query, "locked", &param)) source << param ; else source << config.video_locked ;

    if (path[1]=='v') {
	 source << "&pause=" ;
	 if (URL::getQueryArg(query, "pause", &param)) source << param ; else source << config.video_pause ;
	 source << "&length=" ;
	 if (URL::getQueryArg(query, "length", &param)) source << param ; else source << config.video_length ;
    } else source << "&length=1" ;

  } else if (path.substr(0,6)=="/relay") {

    // Any source

    URL::getQueryArg(query, "src", &param) ;
    source << param ;

  } else if (path.substr(0,6)=="/movie") {

    // Movie file stored somewhere below the server root directory

    std::string filename(path,6,path.length()-6) ;
    filename = reducePath(filename.c_str()) ;
    trimString(filename," ./") ;
    source << "file:" << config.repository << "/" << filename ;

  } else {

    // Default source from videoServer configuration

    source << "videoin:/" << config.video_device << "/" << config.video_node ;
    source << "?size=" << config.video_size << "&locked=" << config.video_locked ;
    source << "&quality=" << config.video_quality << "&pause=" << config.video_pause ;
    source << "&length=" << config.video_length ;

  }

  return source.str() ;
}

// -------------------------------------------------------------------------------

static std::string
describeSource(const vsConfiguration &config, const std::string &source) {
  URL url(source) ;

  if (url.getScheme()!="videoin") return "relay" ;

  std::stringstream description ;

  description << "local " ;
  std::string device=config.video_device, node=config.video_node ;
  std::string size=config.video_size ;
  int length=config.video_length ;
  float pause=config.video_pause ;
  int locked=config.video_locked ;
  int quality=config.video_quality ;
  decodeVideoInURL(url, &device, &node, &size, &length, &pause, &locked, &quality) ;
  description << device << " " << node << " " << locked << " " << size << " " << quality ;
  description << " " << length << " " << pause ;

  return description.str()
 ;
}

// -------------------------------------------------------------------------------

namespace videoServer {

  // -------------------------------------------------------------------------------

  std::string
  vsPreNotifier(const vsConfiguration &config, nucleo::URL &url) {
    std::string request ;

    std::string path = url.getPath() ;
    std::string query = url.getQuery() ;

    if (path.substr(0,5)=="/push") {
	 std::string tmp(path,5,path.length()-5) ;
	 std::string source = parseSource(config, tmp, query) ;
	 std::string description = describeSource(config, source) ;
	 request = "push " + source + "\n" + description + "\n" ;
    } else if (path.substr(0,5)=="/vstp") {
	 std::string host ;
	 URL::getQueryArg(query, "host", &host) ;
	 if (host=="") host = "?" ;
	 std::string port ;
	 URL::getQueryArg(query, "port", &port) ;
	 if (port=="") port = "?" ;
	 std::string tmp(path,5,path.length()-5) ;
	 std::string source = parseSource(config, tmp, query) ;
	 std::string description = describeSource(config, source) ;
	 request = "vstp " + source + " " + host +" " + port + "\n" + description + "\n" ;
    } else if (path.substr(0,5)=="/grab") {
	 std::string source = parseSource(config, "/photo", query) ;
	 std::string description = describeSource(config, source) ;
	 request = "grab " + source + "\n" + description + "\n" ;
    } else if (path.substr(0,8)=="/realize") {
	 std::string tmp(path,8,path.length()-8) ;
	 std::string source = parseSource(config, tmp, query) ;
	 std::string description = describeSource(config, source) ;
	 request = "realize " + source + "\n" + description + "\n" ;
    } else if (path.substr(0,8)=="/mpegify") {
	 std::string tmp(path,8,path.length()-8) ;
	 std::string source = parseSource(config, tmp, query) ;
	 std::string description = describeSource(config, source) ;
	 request = "mpegify " + source + "\n" + description + "\n" ;
    } else {
	 path = reducePath(path.c_str()) ;
	 trimString(path," ./") ;
	 // request = "send " + config.root + "/" + path +"\nsend\n" ;
	 request = "send " + path +"\nsend\n" ;
    }

    return request ;
  }

  // -------------------------------------------------------------------------------

}
