/*
 *
 * apps/videoServer/vsService.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "vsConfiguration.H"
#include "vsSubProcess.H"
#include "vsPreNotifier.H"
#include "vsPostNotifier.H"

#include <nucleo/core/URL.H>
#include <nucleo/network/tcp/TcpConnection.H>
#include <nucleo/network/http/HttpMessage.H>

#include <unistd.h>
#include <sys/stat.h>

#include <iostream>
#include <sstream>

using namespace nucleo ;

namespace videoServer {

  void
  vsServe(const vsConfiguration &config, nucleo::TcpConnection *client) {

    int fd = client->getFd() ;

    // -----------------------------------------------------------------------------------
    // Read the request

    HttpMessage request ;

    if (! request.parseFromStream(fd)) {
	 if (config.debug_level>=1) std::cerr << "Bad request" << std::endl ;
	 char *response = "HTTP/1.0 400 Bad Request\015\012\015\012" ;
	 write(fd, response, strlen(response)) ;
	 return ;
    }

    if (config.debug_level>=1) std::cerr << std::endl ;

    // -----------------------------------------------------------------------------------
    // Get process, user and machine info

    pid_t pid = getpid() ;

    std::string machine ;
    try { machine = client->machineLookUp() ; } catch (...) { machine = "?" ; }

    std::string user ;
    try { user = client->userLookUp() ; } catch (...) { user = "?" ; }

    std::string referer ;
    if (! request.getHeader("referer", referer)) referer = "?" ;

    std::string agent ;
    if (! request.getHeader("user-agent", agent)) agent = "?" ;

    std::stringstream s ;
    s << pid << " " << user << " " << machine << " " << referer << " " << agent ;

    std::string meta_info = s.str() ;

    if (config.debug_level>=1) std::cerr << meta_info << std::endl ;

    // -----------------------------------------------------------------------------------
    // Parse the request

    std::string startline = request.startLine() ;
    std::string method = extractNextWord(startline) ;
    std::string resource = extractNextWord(startline) ;

    std::string url_s = "videoServer:" ;
    url_s.append(resource) ;

    if (config.debug_level>=1) std::cerr << url_s << std::endl ;

    URL url(url_s) ;

    // -----------------------------------------------------------------------------------
    // Get a human-friendly (i.e. simple) description of the request and
    // additional information for control and notification scripts

    std::string pre_notifier = vsPreNotifier(config, url) ;

    if (config.debug_level>=1)
	 std::cerr << "---" << std::endl << pre_notifier << "---" << std::endl ;

    // -----------------------------------------------------------------------------------
    // Trigger control and notification mechanisms

    std::string input = meta_info + "\n" + pre_notifier + "\n" ;

    vsSubProcess *kwai=0 ;
    std::string output ;

    struct stat notifattrs ;
    if (!stat(config.notifier.c_str(), &notifattrs)) { // Should test more than this...
#if 0
	 // "/Users/roussel/casa/lib/videoServer/notifier.pdq"
	 const char *notifier = config.notifier.c_str() ;
	 char * argv[] = {(char *)notifier,0} ;
	 kwai = new vsSubProcess(getenv("SHELL"),argv) ;
#else
	 kwai = new vsSubProcess(config.notifier.c_str()) ;
#endif
	 kwai->send(input) ;
	 output = kwai->receive("\n") ;
    } else {
	 std::cerr << "Warning: can't find " << config.notifier << std::endl ;
	 output = pre_notifier ;
    }

    std::string::size_type p = output.find_first_of("\r\n") ;
    if (p!=std::string::npos) output.erase(p,output.length()-p) ;

    trimString(output) ;

    if (config.debug_level>=1)
	 std::cerr << output << std::endl << "---" << std::endl ;

    // -----------------------------------------------------------------------------------
    // Execute the service (input should be a valid output...)

    vsPostNotifier(config, output, client) ;

    // -----------------------------------------------------------------------------------

    if (kwai) kwai->send("done") ;
    delete kwai ;
  }

}
