/*
 *
 * nucleo/gl/window/glWindow_AGL.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>
#include <nucleo/gl/window/glWindow_AGL.H>
#include <nucleo/core/ReactiveEngine.H>

#include <OpenGL/OpenGL.h>

// fullscreen
#include <QuickTime/Movies.h>

// CGWarpMouseCursorPosition
#include <ApplicationServices/ApplicationServices.h>

#define DEBUG_LEVEL 0

namespace nucleo {

  static inline void
  debugModifiers(std::string msg, UInt32 modifiers) {
    std::cerr << "kEventClassKeyboard: " << msg << " " << modifiers << std::endl ;
    std::cerr << "   cmdKey: " << (modifiers & cmdKey) << std::endl ;
    std::cerr << "   shiftKey: " << (modifiers & shiftKey) << std::endl ;
    std::cerr << "   alphaLock: " << (modifiers & alphaLock) << std::endl ;
    std::cerr << "   optionKey: " << (modifiers & optionKey) << std::endl ;
    std::cerr << "   controlKey: " << (modifiers & controlKey) << std::endl ;
    std::cerr << "   rightShiftKey: " << (modifiers & rightShiftKey) << std::endl ;
    std::cerr << "   rightOptionKey: " << (modifiers & rightOptionKey) << std::endl ;
    std::cerr << "   rightControlKey: " << (modifiers & rightControlKey) << std::endl ;
    
    std::cerr << "   NumLock: " << (modifiers & kEventKeyModifierNumLockMask) << std::endl ;
    std::cerr << "   Fn: " << (modifiers & kEventKeyModifierFnMask) << std::endl ;
  }

  // -----------------------------------------------------------------------

  glWindow
  *createAGLwindow(long options, long eventmask) {
    return new glWindow_AGL(options, eventmask) ;
  }

  // -----------------------------------------------------------------------

  OSStatus
  glWindow_AGL::_windowEventHandler(EventHandlerCallRef nextHandler,
							 EventRef theEvent,
							 void* userData) {
    glWindow_AGL *obj = (glWindow_AGL*) userData ;

    // -- Standard handler ---

    OSStatus result = CallNextEventHandler(nextHandler, theEvent) ;
    if (result!=eventNotHandledErr && result!=noErr) {
	 std::cerr << "glWindow_AGL: next event handler returned " << result << std::endl ;
	 return result ;
    }

    // -----------------------

    UInt32 eclass=GetEventClass(theEvent),
	 ekind=GetEventKind(theEvent) ;

    bool eventHandled = true ;

    switch(eclass) {

    case kEventClassMouse :
	 switch(ekind){

	 case kEventMouseDown :
	 case kEventMouseUp : {
	   Point point;
	   GetEventParameter(theEvent, kEventParamMouseLocation, typeQDPoint, NULL, sizeof(Point), NULL, &point);
	   int x = point.h - obj->_rect.left ; 
	   int y = point.v - obj->_rect.top ;
	   if (FindWindow(point, &(obj->_window)) == inContent) {
		glWindow::event *e = new glWindow::event ;
		e->type = (ekind==kEventMouseDown) ? glWindow::event::buttonPress : glWindow::event::buttonRelease;
		e->x = x ;
		e->y = y ; 
		EventMouseButton mouseButton ;
		GetEventParameter(theEvent, kEventParamMouseButton, typeMouseButton, NULL, sizeof(EventMouseButton), NULL, &mouseButton) ;
		if (mouseButton!=1)
		  e->button = mouseButton ;
		else {
		  if (obj->_modifiers&controlKey) e->button = 3 ;
		  else if (obj->_modifiers&optionKey) e->button = 2 ;
		  else e->button = 1 ;
		}
		obj->_equeue.push(e) ;
	   }
	 } break;          

	 case kEventMouseMoved :
	 case kEventMouseDragged : {
	   Point point;
	   GetEventParameter(theEvent, kEventParamMouseLocation, typeQDPoint, NULL, sizeof(Point), NULL, &point);
	   int x = point.h - obj->_rect.left ; 
	   int y = point.v - obj->_rect.top ;
	   if (FindWindow(point, &(obj->_window)) == inContent) {
		glWindow::event *e = new glWindow::event ;
		e->type = glWindow::event::pointerMotion ;
		e->x = x ;
		e->y = y ;
		obj->_equeue.push(e) ;
	   }
	 } break;

	 case kEventMouseWheelMoved : {
	   EventMouseWheelAxis axis ;
	   GetEventParameter(theEvent, kEventParamMouseWheelAxis, typeMouseWheelAxis, NULL, sizeof(axis), NULL, &axis ) ;

	   SInt32 delta ;
	   GetEventParameter(theEvent, kEventParamMouseWheelDelta, typeSInt32, NULL, sizeof(delta), NULL, &delta ) ;

	   glWindow::event *e = new glWindow::event ;
	   e->type = glWindow::event::wheelMotion ;
	   e->axis = (int)axis ;
	   e->delta = (int)delta ;
	   obj->_equeue.push(e) ;
	 } break ;

	 default :
	   eventHandled = false ;
	   break ;
	 }
	   break;

	 case kEventClassKeyboard :
	   switch(ekind){
	   case kEventRawKeyDown :
	   case kEventRawKeyRepeat:
	   case kEventRawKeyUp : {
		glWindow::event *e = new glWindow::event ;

		if (ekind==kEventRawKeyDown)
		  e->type = glWindow::event::keyPress ;
		else // Up or Repeat
		  e->type = glWindow::event::keyRelease ;

		char charCode ;
		GetEventParameter (theEvent, kEventParamKeyMacCharCodes, typeChar, NULL, sizeof(char), NULL, &charCode);

		UInt32 keyCode ;
		GetEventParameter(theEvent, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &keyCode);

		bool unknown_key = false ;

		switch(charCode) {
		case 16:
		  switch(keyCode) {
		  case 122: e->keysym = XK_F1 ; e->keystr = "<F1>" ; break ;
		  case 120: e->keysym = XK_F2 ; e->keystr = "<F2>" ; break ;
		  case 99: e->keysym = XK_F3 ; e->keystr = "<F3>" ; break ;
		  case 118: e->keysym = XK_F4 ; e->keystr = "<F4>" ; break ;
		  case 96: e->keysym = XK_F5 ; e->keystr = "<F5>" ; break ;
		  case 97: e->keysym = XK_F6 ; e->keystr = "<F6>" ; break ;
		  case 98: e->keysym = XK_F7 ; e->keystr = "<F7>" ; break ;
		  case 100: e->keysym = XK_F8 ; e->keystr = "<F8>" ; break ;
		  case 101: e->keysym = XK_F9 ; e->keystr = "<F9>" ; break ;
		  case 109: e->keysym = XK_F10 ; e->keystr = "<F10>" ; break ;
		  case 103: e->keysym = XK_F11 ; e->keystr = "<F11>" ; break ;
		  case 111: e->keysym = XK_F12 ; e->keystr = "<F12>" ; break ;
		  default: unknown_key = true ; break ;
		  }
		  break ;
#define KEYTRANS(cc,sym,str) case cc: e->keysym=sym; e->keystr=str ; break ;
#include <nucleo/gl/window/_macos2keysym.h>
#undef KEYTRANS
		default:
		  unknown_key = true ;
		  break;
		}

		if (unknown_key) {
		  std::cerr << "glWindow_AGL: unknown MacOS key (c=" << (int)charCode << ",k=" << keyCode << ")" << std::endl;
		} else {
		  obj->_equeue.push(e) ;
		  if (ekind==kEventRawKeyRepeat) {
		    glWindow::event *e2 = new glWindow::event ;
		    e2->type = glWindow::event::keyPress ;
		    e2->keysym = e->keysym ;
		    e2->keystr = e->keystr ;
		    obj->_equeue.push(e2) ;
		  }
		}
	   
	   } break;

	   case kEventRawKeyModifiersChanged: {
		UInt32 m ;
		GetEventParameter(theEvent, kEventParamKeyModifiers,typeUInt32,NULL,sizeof(UInt32),NULL,&m) ;

		UInt32 mdelta = m^obj->_modifiers ;
		// debugModifiers("modifiers = ",m) ;
		obj->_modifiers = m ;

		if (mdelta & cmdKey) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&cmdKey) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Meta_L ;
		  e->keystr = "<LeftMeta>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & shiftKey) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&shiftKey) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Shift_L ;
		  e->keystr = "<LeftShift>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & rightShiftKey) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&rightShiftKey) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Shift_R ;
		  e->keystr = "<RightShift>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & controlKey) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&controlKey) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Control_L ;
		  e->keystr = "<LeftControl>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & rightControlKey) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&rightControlKey) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Control_R ;
		  e->keystr = "<RightControl>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & optionKey) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&optionKey) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Alt_L ;
		  e->keystr = "<LeftAlt>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & rightOptionKey) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&rightOptionKey) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Alt_R ;
		  e->keystr = "<RightAlt>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & kEventKeyModifierFnMask) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&kEventKeyModifierFnMask) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Super_L ;
		  e->keystr = "<LeftSuper>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & alphaLock) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&alphaLock) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Caps_Lock ;
		  e->keystr = "<CapsLock>" ;
		  obj->_equeue.push(e) ;
		} else if (mdelta & kEventKeyModifierNumLockMask) {
		  glWindow::event *e = new glWindow::event ;
		  e->type = (m&kEventKeyModifierNumLockMask) ? event::keyPress : event::keyRelease ;
		  e->keysym = XK_Num_Lock ;
		  e->keystr = "<NumLock>" ;
		  obj->_equeue.push(e) ;
		}
	   } break ;
	   default :
		eventHandled = false ;
		break;
	   }
	   break ;

	 case kEventClassWindow :
	   switch(ekind){
	   case kEventWindowClose : {
		glWindow::event *e = new glWindow::event ;
		e->type = glWindow::event::destroy;
		obj->_equeue.push(e) ;
	   } break ;
	   case kEventWindowDrawContent : {
		glWindow::event *e = new glWindow::event ;
		e->type = glWindow::event::expose ;
		obj->_equeue.push(e) ;
	   } break;          
	   case kEventWindowBoundsChanged : {
		UInt32 attributes = 0 ;
		GetEventParameter(theEvent, kEventParamAttributes, typeUInt32, 
								   NULL, sizeof(UInt32), NULL, &attributes) ;
		bool resizing = (attributes & kWindowBoundsChangeSizeChanged) ;

		GetWindowBounds(obj->_window, kWindowContentRgn, &obj->_rect) ;
		unsigned int width = obj->_rect.right - obj->_rect.left ;
		unsigned int height = obj->_rect.bottom - obj->_rect.top;

		if (resizing && obj->_aspect) {
		  double a = (double)width/(double)height ;
		  if (a!=obj->_aspect) {
		    width = (unsigned int)(height*obj->_aspect) ;
		    SizeWindow(obj->_window, width, height, false) ;
		  }
		}

		aglSetCurrentContext(obj->_glContext) ;
		aglUpdateContext(obj->_glContext) ;

		glWindow::event *e = new glWindow::event ;
		e->type = glWindow::event::configure ;
		e->x = obj->_rect.left ;
		e->y = obj->_rect.top ;
		e->width = width ;
		e->height = height ;
		obj->_equeue.push(e) ;

		// std::cerr << "glWindow_AGL event: " << resizing << " " ; e->debug(std::cerr) ; std::cerr << std::endl ;

	   } break;
	   default:
		eventHandled = false ;
		break ;
	   }
	   break;	 

	 default :
	   eventHandled = false ;
	   break ;
	 }

	 if (eventHandled) {
	   obj->notifyObservers() ;
	   return noErr ;
	 }

	 char *ptr = (char *)&eclass ;
	 std::cerr << "unexpected event: " ;
	 std::cerr << "'" << ptr[0] << ptr[1] << ptr[2] << ptr[3] << "' (" << eclass << ")" ;
	 std::cerr << " / " << ekind << std::endl ;
	 return eventNotHandledErr ;
    }

    // -----------------------------------------------------------------------

    glWindow_AGL::glWindow_AGL(long options, long eventmask)
	 : _fullscreenRestoreState(0) {
	 debugEvents = false ;
	 _aspect = 0.0 ;
	 _modifiers = 0 ;
	 _mapped = false ;
	 _options = options ;
	 _eventmask = eventmask ;
	 SetRect(&_rect, 50, 50, 370, 290) ;

	 // Some might like kWindowMetalAttribute
	 // This one might be useful someday: kWindowIgnoreClicksAttribute
	 // |kWindowStandardHandlerAttribute
	 CreateNewWindow(kDocumentWindowClass,
				  kWindowAsyncDragAttribute|kWindowStandardDocumentAttributes,
				  &_rect, &_window) ;

	 setup(_options, _eventmask) ;

	 // So that we can notify the observers after the constructor has
	 // returned...
	 ReactiveEngine::notify(this, this) ; 
    }
      
    glWindow_AGL::~glWindow_AGL(void) {
	 aglSetCurrentContext(NULL);
	 aglSetDrawable(_glContext, NULL);
	 aglDestroyContext(_glContext);
	 ReleaseWindow(_window);
    }
    
    void
	 glWindow_AGL::setup(long options, long eventmask) {                        
	 AGLPixelFormat apf = 0;      

	 GLint attributeList[] = { AGL_RGBA,
						  AGL_RED_SIZE, 8,
						  AGL_GREEN_SIZE, 8,
						  AGL_BLUE_SIZE, 8,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
						  AGL_NONE,
	 };

	 int attributeListSize = 7 ;

	 if (options&glWindow::DOUBLE_BUFFER)
	   attributeList[attributeListSize++] = AGL_DOUBLEBUFFER ;

	 if (options&glWindow::STENCIL) {
	   attributeList[attributeListSize++] = AGL_STENCIL_SIZE ;
	   attributeList[attributeListSize++] = 8 ;
	 }

	 int idepth=0 ;
	 if (options&glWindow::DEPTH) {
	   attributeList[attributeListSize++] = AGL_DEPTH_SIZE ;
	   idepth = attributeListSize ;
	   attributeList[attributeListSize++] = 24 ;
	 }

	 int depth, adepth=0, sizeBeforeAlpha=attributeListSize ;

	 if (options&glWindow::ALPHA) {
	   attributeListSize+=2 ;
	   adepth=8 ;
	 }

	 GDHandle device = GetMainDevice();
	 do {
	   if (options&glWindow::ALPHA) {
		attributeList[sizeBeforeAlpha] = AGL_ALPHA_SIZE ;
		adepth = attributeList[sizeBeforeAlpha+1] = adepth ;
	   }

	   for (depth=8; depth; depth--) {
		attributeList[2] = attributeList[4] = attributeList[6] = depth ;
		apf = aglChoosePixelFormat(&device, 1, attributeList); // NULL, 0, attributeList);
		if (apf) break ;
	   }
	   if (apf) break ;

	   adepth-- ;

	 } while (adepth>=0) ;

	 if (!apf) {
	   std::cerr << "glWindow_AGL: No suitable apf" << std::endl ;
	   exit(1) ;
	 }
      
#if DEBUG_LEVEL>=2
	 std::cerr << "glWindow_AGL: rgb-depth=" << depth  ;
	 if (options&glWindow::DEPTH) std::cerr << " depth=" << attributeList[idepth] ;
	 std::cerr << " alpha=" << adepth ;
	 std::cerr << " apf=" << std::hex << apf << std::dec << std::endl ;
#elif DEBUG_LEVEL>=1
	 if ((options&glWindow::ALPHA) && !adepth) std::cerr << "glWindow_AGL warning: no alpha plane available ! (depth=" << depth << ")" << std::endl ;
#endif

	 _glContext = aglCreateContext(apf, NULL); // NULL stands for other AGLContext with which to share display lists
	 aglDestroyPixelFormat(apf);
	 aglSetDrawable(_glContext, GetWindowPort(_window));
   
	 // Events --------------

	 unsigned int lastEventType = 0;

	 _macEvents[lastEventType].eventClass = kEventClassWindow ;
	 _macEvents[lastEventType++].eventKind = kEventWindowBoundsChanged ;

	 if (eventmask & glWindow::event::pointerMotion){
	   _macEvents[lastEventType].eventClass = kEventClassMouse ;
	   _macEvents[lastEventType++].eventKind = kEventMouseMoved ;
	   _macEvents[lastEventType].eventClass = kEventClassMouse ;
	   _macEvents[lastEventType++].eventKind = kEventMouseDragged ;
	 }
	 if (eventmask & glWindow::event::wheelMotion){
	   _macEvents[lastEventType].eventClass = kEventClassMouse ;
	   _macEvents[lastEventType++].eventKind = kEventMouseWheelMoved ;
	 }
	 if (eventmask & glWindow::event::buttonPress){
	   _macEvents[lastEventType].eventClass = kEventClassMouse ;
	   _macEvents[lastEventType++].eventKind = kEventMouseDown ;
	 }
	 if (eventmask & glWindow::event::buttonRelease){
	   _macEvents[lastEventType].eventClass = kEventClassMouse ;
	   _macEvents[lastEventType++].eventKind = kEventMouseUp ;
	 }

	 if (eventmask & glWindow::event::keyPress){
	   _macEvents[lastEventType].eventClass = kEventClassKeyboard ;
	   _macEvents[lastEventType++].eventKind = kEventRawKeyDown ;
	 }
	 if (eventmask & glWindow::event::keyRelease){
	   _macEvents[lastEventType].eventClass = kEventClassKeyboard ;
	   _macEvents[lastEventType++].eventKind = kEventRawKeyUp ;
	 }
	 if (eventmask & (glWindow::event::keyPress|glWindow::event::keyRelease)){
	   _macEvents[lastEventType].eventClass = kEventClassKeyboard ;
	   _macEvents[lastEventType++].eventKind = kEventRawKeyModifiersChanged ;
	   _macEvents[lastEventType].eventClass = kEventClassKeyboard ;
	   _macEvents[lastEventType++].eventKind = kEventRawKeyRepeat ;
	 }

	 if (eventmask & glWindow::event::expose){
	   _macEvents[lastEventType].eventClass = kEventClassWindow ;
	   _macEvents[lastEventType++].eventKind = kEventWindowDrawContent ;
	 }

	 if (eventmask & glWindow::event::destroy) {
	   _macEvents[lastEventType].eventClass = kEventClassWindow ;
	   _macEvents[lastEventType++].eventKind = kEventWindowClose ;
	 }

	 if (eventmask & glWindow::event::enter)
	   std::cerr << "glWindow_AGL: 'enter' events are not supported..." << std::endl ;
	 if (eventmask & glWindow::event::leave)
	   std::cerr << "glWindow_AGL: 'leave' events are not supported..." << std::endl ;

	 _nbMacEvents = lastEventType;

	 InstallStandardEventHandler(GetWindowEventTarget(_window)) ;
	 InstallStandardEventHandler(GetApplicationEventTarget()) ;
	 InstallEventHandler(GetWindowEventTarget(_window),
					 NewEventHandlerUPP(glWindow_AGL::_windowEventHandler),
					 _nbMacEvents, _macEvents,
					 (void *)this,
					 NULL /* outRef */) ;

	 unsigned int width, height ;
	 int x, y ;
	 getGeometry(&width, &height, &x, &y) ;
	 glWindow::event *e = new glWindow::event ;
	 e->type = glWindow::event::configure ;
	 e->width = width ;
	 e->height = height ;
	 e->x = x ;
	 e->y = y ;
	 _equeue.push(e) ;
    }

    // -----------------------------------------------------------------------

    void
	 glWindow_AGL::showCursor(bool activate) {
	 if (activate) ShowCursor() ; else HideCursor() ;
    }

    void
	 glWindow_AGL::warpCursor(int x, int y) {
	 CGPoint newCursorPosition ;
	 newCursorPosition.x = _rect.left + x ;
	 newCursorPosition.y = _rect.top + y ;
	 // std::cerr << "newCursorPosition: (" << newCursorPosition.x << ", " << newCursorPosition.y << ")" << std::endl ;
	 /*CGEventErr err = */ CGWarpMouseCursorPosition(newCursorPosition) ;
	 // std::cerr << "CGWarpMouseCursorPosition returned " << err << std::endl ;
    }

    void
	 glWindow_AGL::syncToVbl(int nb_vbls) {
	 // do we need this: CGLFlushDrawable(CGLGetCurrentContext()) ;
	 aglSetCurrentContext(_glContext) ;      
	 long params[] = { nb_vbls } ; // number of VBLs to wait for
	 CGLSetParameter(CGLGetCurrentContext(), kCGLCPSwapInterval, params) ;
    }

    void
	 glWindow_AGL::raise(void) {
	 SelectWindow(_window) ;
    }

    // -----------------------------------------------------------------------

    void
	 glWindow_AGL::map(void) {
	 if (!_mapped) {
	   ShowWindow(_window) ;
	   _mapped = true ;
	 }
    }
    
    void
	 glWindow_AGL::unmap(void) {
	 if (_mapped) {
	   HideWindow(_window);
	   _mapped = false;
	 }
    }

    void
	 glWindow_AGL::setTitle(const char *title) {
	 Str255 sTitle;
	 CopyCStringToPascal(title, sTitle);
	 SetWTitle(_window, sTitle);
    }

    void
	 glWindow_AGL::getGeometry(unsigned int *width, unsigned int *height, int *x, int *y) {
	 Rect rect;
	 GetWindowBounds(_window, kWindowContentRgn, &rect);
	 if(x) *x = rect.left; if(y) *y = rect.top;
	 if(width) *width = rect.right - rect.left;
	 if(height) *height = rect.bottom - rect.top;
    }
    
    void 
	 glWindow_AGL::setGeometry(unsigned int width, unsigned int height, int x, int y) {
	 if (!_mapped) map() ;
	 SizeWindow(_window, width, height, false);
	 MoveWindow(_window, x, y, false);
    }

    void
	 glWindow_AGL::fullScreenMode(bool activate) {
	 // std::cerr << "glWindow_AGL: warning: 'fullScreenMode' creates a new GL context..." << std::endl ;

	 if(activate) {
	   unmap();
	   aglSetCurrentContext(NULL);
	   aglSetDrawable(_glContext, NULL);
	   aglDestroyContext(_glContext);
	   ReleaseWindow(_window);

	   long flags = 0; //fullScreenHideCursor;
	   BeginFullScreen( &_fullscreenRestoreState, NULL, 0, 0, &_window, NULL, flags);

	   // This will push a configure event
	   setup(_options, _eventmask);

	   // Bring the window to the front
	   SelectWindow(_window);

	   glWindow::event *e = new glWindow::event ;
	   e->type = glWindow::event::expose ;
	   _equeue.push(e) ;
	 } else if(_fullscreenRestoreState) {
	   aglSetCurrentContext(NULL);
	   aglSetDrawable(_glContext, NULL);
	   aglDestroyContext(_glContext);
	   ReleaseWindow(_window);

	   EndFullScreen( _fullscreenRestoreState, NULL);
	   _fullscreenRestoreState=0;

	   SetRect(&_rect, 50, 50, 370, 290) ;
	   CreateNewWindow(kDocumentWindowClass,
				    kWindowStandardDocumentAttributes | kWindowStandardHandlerAttribute,
				    &_rect, &_window) ;

	   // This will push a configure event
	   setup(_options, _eventmask) ;

	   if(_mapped) ShowWindow(_window) ;
	 }
    }

    void
	 glWindow_AGL::setAspectRatio(int width, int height) {
	 _aspect = (double)width/(double)height ;
    }

    // -----------------------------------------------------------------------
    
    void
	 glWindow_AGL::react(Observable*) {
	 // std::cerr << "glWindow_AGL::react" << std::endl ;
	 notifyObservers() ;
    }

    bool
	 glWindow_AGL::getNextEvent(event *e) {
	 if (_equeue.empty()) return false ;

	 event *eq = _equeue.front() ;
	 _equeue.pop() ;
	 *e = *eq ;
	 if (debugEvents) {
	   std::cerr << "glWindow_AGL event: " ;
	   e->debug(std::cerr) ;
	   std::cerr << std::endl ;
	 }
	 delete eq ;

	 return true ;
    }

    // -----------------------------------------------------------------------

  }
