/*
 *
 * nucleo/gl/window/glWindow_GLX.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/core/ReactiveEngine.H>
#include <nucleo/gl/window/glWindow_GLX.H>

#if HAVE_X11_XM_MWMUTIL_H
#  include <X11/Xm/MwmUtil.h>
#elif HAVE_XM_MWMUTIL_H
#  include <Xm/MwmUtil.h>
#endif

#include <X11/cursorfont.h>
#include <X11/Xatom.h>

#include <iostream>
#include <stdexcept>

#define DEBUG_LEVEL 0

namespace nucleo {

  // -----------------------------------------------------------------------

  glWindow
  *createGLXwindow(long options, long eventmask) {
    return new glWindow_GLX(options, eventmask) ;
  }

  // ------------------------------------------------------------------

  unsigned int
  glWindow_GLX::getScreenWidth() {
    return DisplayWidth(_xDisplay,DefaultScreen(_xDisplay)) ;
  }

  unsigned int
  glWindow_GLX::getScreenHeight() {
    return DisplayHeight(_xDisplay,DefaultScreen(_xDisplay)) ;
  }

  // ------------------------------------------------------------------

  glWindow_GLX::glWindow_GLX(Display *disp, Window parent,
					    long options, long eventmask) {
    debugEvents = false ;
    _xDisplay = disp ;
    _xParent = parent ;
    setup(options, eventmask) ;

    // So that we can notify the observers after the constructor has
    // returned...
    ReactiveEngine::notify(this, this) ; 
  }

  // ------------------------------------------------------------------

  glWindow_GLX::glWindow_GLX(long options, long eventmask) {
    debugEvents = false ;
    _xDisplay = XOpenDisplay(0) ;
    _xParent = 0 ;
    if (!_xDisplay) throw std::runtime_error("glWindow_GLX: can't open display") ;
    setup(options, eventmask) ;

    // So that we can notify the observers after the constructor has
    // returned...
    ReactiveEngine::notify(this, this) ; 
  }

  // ------------------------------------------------------------------

  void
  glWindow_GLX::setup(long options, long eventmask) {
    _mapped = false ;

    _fk = FileKeeper::create(ConnectionNumber(_xDisplay), FileKeeper::R);
    subscribeTo(_fk) ;

    XVisualInfo *vi = 0 ;
    XSetWindowAttributes swa ;
    int attributeList[] = {
	 GLX_RGBA,
	 GLX_RED_SIZE, 8,
	 GLX_GREEN_SIZE, 8,
	 GLX_BLUE_SIZE, 8,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None,
	 (int)None
    } ;
    int attributeListSize = 7 ;

    if (options&glWindow::DOUBLE_BUFFER) attributeList[attributeListSize++] = GLX_DOUBLEBUFFER ;

    if (options&glWindow::STENCIL) {
	 attributeList[attributeListSize++] = GLX_STENCIL_SIZE ;
	 attributeList[attributeListSize++] = 8 ;
    }

    int idepth=0 ;
    int ddepth = 0;
    if (options&glWindow::DEPTH) {
	 attributeList[attributeListSize++] = GLX_DEPTH_SIZE ;
	 idepth = attributeListSize ;
	 // was 1 (changed 25 apr 2002)
	 // 1 is tried if glXChooseVisual fail (2004-02-22)
	 ddepth = attributeList[attributeListSize++] = 24 ;
    }

    int depth, adepth=0, sizeBeforeAlpha=attributeListSize ;

    if (options&glWindow::ALPHA) {
	 attributeListSize+=2 ;
	 adepth=8 ;
    }

    do {
      do {

	   if (options&glWindow::ALPHA) {
		attributeList[sizeBeforeAlpha] = GLX_ALPHA_SIZE ;
		adepth = attributeList[sizeBeforeAlpha+1] = adepth ;
	   }

	   for (depth=8; depth; depth--) {
		attributeList[2] = attributeList[4] = attributeList[6] = depth ;
		vi = glXChooseVisual(_xDisplay, DefaultScreen(_xDisplay), attributeList);
		if (vi)
		  break ;
	   }
	   if (vi) break ;

	   adepth-- ;

      } while (adepth>=0) ;

      if (vi) break ;

      if ((options&glWindow::DEPTH) && ddepth == 24) {
	   attributeList[idepth] = 1 ;
	   ddepth = 1;
      } else { ddepth = 0; }

    } while (ddepth) ;

    if (!vi) {
      std::cerr << "glWindow_GLX : No suitable visual" << std::endl ;
      exit(1) ;
    }

#if DEBUG_LEVEL>=2
    std::cerr << "glWindow_GLX: rgb-depth=" << depth  ;
    if (options&glWindow::DEPTH) std::cerr << " depth=" << attributeList[idepth] ;
    std::cerr << " alpha=" << adepth ;
    std::cerr << " visual=0x" << hex << vi->visualid << dec << " (compare with 'glxinfo -b')" << std::endl ;
#else
#  if DEBUG_LEVEL>=1
    if ((options&glWindow::ALPHA) && !adepth) std::cerr << "glWindow_GLX warning: no alpha plane available ! (depth=" << depth << ")" << std::endl ;
#  endif
#endif

    _glContext = glXCreateContext(_xDisplay, vi, 0, GL_TRUE);

    if (!_xParent) _xParent = RootWindow(_xDisplay, vi->screen) ;
    int x=0, y=0 ;
    unsigned int w=1, h=1 ;
    unsigned long valuemask = 0 ;

    // swa.override_redirect = False ;       valuemask |= CWOverrideRedirect ;
    swa.backing_store = NotUseful ;       valuemask |= CWBackingStore ;
    swa.background_pixel = 0xFFFFFFFF ;   valuemask |= CWBackPixel ;
    swa.save_under = False ;              valuemask |= CWSaveUnder ;
    swa.border_pixel = 0xFFFFFFFF ;       valuemask |= CWBorderPixel ;
    
    swa.colormap = XCreateColormap(_xDisplay, RootWindow(_xDisplay, vi->screen), vi->visual, AllocNone);
    valuemask |= CWColormap ;

    swa.event_mask = 0 ;
    if (eventmask & glWindow::event::configure) swa.event_mask |= StructureNotifyMask ;
    if (eventmask & glWindow::event::expose) swa.event_mask |= ExposureMask ;
    if (eventmask & glWindow::event::destroy) swa.event_mask |= StructureNotifyMask ;
    if (eventmask & glWindow::event::enter) swa.event_mask |= EnterWindowMask ;
    if (eventmask & glWindow::event::leave) swa.event_mask |= LeaveWindowMask ;
    if (eventmask & glWindow::event::keyPress) swa.event_mask |= KeyPressMask ;
    if (eventmask & glWindow::event::keyRelease) swa.event_mask |= KeyReleaseMask ;
    if (eventmask & glWindow::event::buttonPress) swa.event_mask |= ButtonPressMask ;
    if (eventmask & glWindow::event::buttonRelease) swa.event_mask |= ButtonReleaseMask ;
    if (eventmask & glWindow::event::pointerMotion) swa.event_mask |= PointerMotionMask ;
    valuemask |= CWEventMask ;

    _xWindow = XCreateWindow(_xDisplay, _xParent,
					    x, y, w, h,
					    0, vi->depth, InputOutput, vi->visual,
					    valuemask, &swa);

    if (eventmask & glWindow::event::destroy) {
      static Atom wmDeleteWindow = XInternAtom(_xDisplay, "WM_DELETE_WINDOW", False);
      XSetWMProtocols(_xDisplay, _xWindow, &wmDeleteWindow, 1);
      // NB: to remove it: XSetWMProtocols(_xDisplay, _xWindow, &wmDeleteWindow, 0);
    }
    
    XFree(vi);

  }

  // ------------------------------------------------------------------

  glWindow_GLX::~glWindow_GLX(void) {
    unsubscribeFrom(_fk) ;
    delete _fk ;
    unmap() ;
    XSync(_xDisplay, False) ;
    glXDestroyContext(_xDisplay, _glContext) ;
    XSync(_xDisplay, False) ;
    XDestroyWindow(_xDisplay, _xWindow) ;
  }

  // ------------------------------------------------------------------

  void
  glWindow_GLX::map(void) {
    if (!_mapped) {
	 XMapWindow(_xDisplay, _xWindow) ;
	 _mapped = true ;
    }
  }

  void
  glWindow_GLX::unmap(void) {
    if (_mapped) {
	 XUnmapWindow(_xDisplay, _xWindow) ;
	 _mapped = false ;
    }
  }

  // ------------------------------------------------------------------

  // use the EWMH spec (see http://www.freedesktop.org/Standards/wm-spec)
  // for certain opertaions. Most "modern" X wm implement this spec (e.g.,
  // FVWM-2.5.x, metacity, kwin, icewm, latest E16, swafish ...etc.)

  void
  *glWindow_GLX::_getAtom(Window id, Atom to_get, Atom type, unsigned int *size)
  {
    unsigned char *retval;
    Atom  type_ret;
    unsigned long  bytes_after, num_ret;
    long length;
    int  format_ret;
    void *data;
    int ok;

    retval = NULL;
    length = 0x7fffffff;
    ok = XGetWindowProperty(_xDisplay, id, to_get, 0, length, False, type,
					   &type_ret, &format_ret, &num_ret, &bytes_after, &retval);

    if ((ok == Success) && (retval) && (num_ret > 0) && (format_ret > 0)) {
      data = malloc(num_ret * (format_ret >> 3));
      if (data) {
	   memcpy(data, retval, num_ret * (format_ret >> 3));
      }
      XFree(retval);
      *size = num_ret * (format_ret >> 3);
      return data;
    }
    if (retval)
      XFree(retval);
    return NULL;
  }

  // check if the current wm is EWMH compliant and support the netProprety
  // property
  bool
  glWindow_GLX::_checkNetSupported(Atom netProprety) {
    Atom wmWindowAtom;
    Atom netSupportedAtom;

    wmWindowAtom = XInternAtom(_xDisplay,"_NET_SUPPORTING_WM_CHECK", False);
    netSupportedAtom = XInternAtom(_xDisplay,"_NET_SUPPORTED", False);
     
    // first we check that the current window manager is EWMH compliant
    // Note that the wm may change during the life of the window
    unsigned int size;
    Window *wmWindow = (Window *)_getAtom(DefaultRootWindow(_xDisplay),
								  wmWindowAtom, XA_WINDOW, &size);
    if (!wmWindow)
      return false;

    // check that the window exist ...
    Window junkRoot;
    int junkX, junkY;
    unsigned int junkWidth, junkHeight, junkBW, junkDepth;
    if (!XGetGeometry(_xDisplay, wmWindow[0], &junkRoot, &junkX, &junkY,
				  &junkWidth, &junkHeight, &junkBW,  &junkDepth)) {
      return false;
    }

    // check that the window is the good corresponding window 
    Window *checkWindow = 0;
    checkWindow = (Window *)_getAtom(wmWindow[0], wmWindowAtom, XA_WINDOW,
							  &size);

    if (!checkWindow || checkWindow[0] != wmWindow[0]) {
      if (checkWindow)
	   free(checkWindow);
      free(wmWindow);
      return false;
    }

    free(wmWindow);
    free(checkWindow);

    // get what the wm support
    Atom *supported;
    supported = (Atom *)_getAtom(DefaultRootWindow(_xDisplay),
						   netSupportedAtom, XA_ATOM, &size);
    if (!supported)
      return false;

    unsigned int num = 0;
    bool found = false;
    while(num < size && !found) {
      if (supported[num] == netProprety) {
	   found = true;
      }
      num++;
    }
    free(supported);
    return found;
  }

  bool
  glWindow_GLX::_ewmhFullScreenMode(bool activate) {
    Atom fullScreenAtom = XInternAtom(_xDisplay,"_NET_WM_STATE_FULLSCREEN",0) ;

    if (!_checkNetSupported(fullScreenAtom))
      return false;
    Atom state = XInternAtom(_xDisplay,"_NET_WM_STATE",0);
    if (!_mapped && activate) {
      XChangeProperty(_xDisplay, _xWindow, state, XA_ATOM, 32, PropModeAppend,
				  (unsigned char*)&fullScreenAtom, 1);
      setGeometry(getScreenWidth(),getScreenHeight(),0,0) ;
    } else if (_mapped) {
      std::cerr << "Ewmh fullscreen " << activate << std::endl;
      XClientMessageEvent ev;
      ev.type = ClientMessage;
      ev.window = _xWindow;
      ev.message_type = state;
      ev.format = 32;
      ev.data.l[0] = (int)activate;
      ev.data.l[1] = fullScreenAtom;
      ev.data.l[2] = 0;
      XSendEvent(_xDisplay, DefaultRootWindow(_xDisplay), False,
			  (SubstructureNotifyMask|SubstructureRedirectMask),
			  (XEvent *) &ev);
      /* the wm move-resize for us. But in the case where we _start_
       * fullscreen and we ask to desactivate the wm cannot know the size
       * we want, so in this case the caller should resize the window
       * itself to the desired size. */
    }
    else {
      // !_mapped && !activate, nothing to do!?
    }
    return true;
  }

  // ------------------------------------------------------------------

  void 
  glWindow_GLX::setTitle(const char *title) {
    XStoreName(_xDisplay, _xWindow, title) ;
  }
  
  void
  glWindow_GLX::getGeometry(unsigned int *width, unsigned int *height,
					   int *x, int *y) {
    int rx, ry ;
    unsigned int rw, rh ;
    Window wDummy ;
    unsigned int uiDummy ;
    XGetGeometry(_xDisplay, _xWindow, &wDummy,
			  &rx, &ry, &rw, &rh,
			  &uiDummy, &uiDummy) ;
    if (x) *x = rx ; if (y) *y = ry ;
    if (width) *width = rw ; if (height) *height = rh ;
  }

  void
  glWindow_GLX::setGeometry(unsigned int width, unsigned int height,
					   int x, int y) {
    if (x<0 || y<0) {
	 Window wDummy ;
	 unsigned int uiDummy, pWidth, pHeight ;
	 int iDummy ;
	 XGetGeometry(_xDisplay, _xParent, &wDummy,
			    &iDummy, &iDummy, &pWidth, &pHeight, &uiDummy, &uiDummy) ;
	 if (x<0) x += pWidth ;
	 if (y<0) y += pHeight ;
    }
    map() ;
    XMoveResizeWindow(_xDisplay, _xWindow, x, y, width, height) ;
  }

  void
  glWindow_GLX::setAspectRatio(int width, int height) {
    XSizeHints *h = XAllocSizeHints() ;
    h->flags = PAspect ;
    h->min_aspect.x = h->max_aspect.x = width ;
    h->min_aspect.y = h->max_aspect.y = height ;
    XSetWMNormalHints(_xDisplay, _xWindow, h) ;
    XFree(h) ;
  }

  void 
  glWindow_GLX::setMinMaxSize(int minwidth, int minheight, int maxwidth, int maxheight) {
    XSizeHints *h = XAllocSizeHints() ;
    h->flags = 0 ;
    if (minwidth>-1 && minheight>-1) {
	 h->flags = h->flags | PMinSize ;
	 h->min_width = minwidth ;
	 h->min_height = minheight ;
    }
    if (maxwidth>-1 && maxheight>-1) {
	 h->flags = h->flags | PMaxSize ;
	 h->max_width = maxwidth ;
	 h->max_height = maxheight ;
    }
    XSetWMNormalHints(_xDisplay, _xWindow, h) ;
    XFree(h) ;
  }

  void
  glWindow_GLX::fullScreenMode(bool activate) {
    if (!_ewmhFullScreenMode(activate)) {
      // EWMH method fails
#ifdef MWM_HINTS_DECORATIONS
      XGrabServer(_xDisplay) ;
      if (_mapped) unmap() ;
      Atom MotifHints = XInternAtom(_xDisplay,"_MOTIF_WM_HINTS",0) ;
      MotifWmHints hints ;
      hints.flags = MWM_HINTS_DECORATIONS ;
      hints.decorations = activate ? 0 : MWM_DECOR_ALL ;
      XChangeProperty(_xDisplay, _xWindow,
				  MotifHints, MotifHints, 32,
				  PropModeReplace, (unsigned char *) &hints, 4) ;
      XUngrabServer(_xDisplay) ;
#endif
      if (activate)
	   setGeometry(getScreenWidth(),getScreenHeight(),0,0) ;
      else
	   {
		// I do not think this is correct; either we should restore the
		// size and position before fullscreen state or we have been
		// started fullscreen and the caller should resize itself.
		// (at least this is the ewmh logic).
		setGeometry(getScreenWidth()/2,getScreenHeight()/2,50,50) ;
	   }
    }
    map() ;
  }

  void
  glWindow_GLX::showCursor(bool activate) {
    Cursor cursor = 0 ;
    if (!activate) {
	 XColor xcf, xcb ;
	 static char m [] = { 0x0 } ;
	 Pixmap pixmap = XCreateBitmapFromData (_xDisplay, _xWindow, m, 1, 1) ;
	 cursor = XCreatePixmapCursor (_xDisplay, pixmap, pixmap, &xcf, &xcb, 0, 0) ;
    } else {
	 cursor = XCreateFontCursor(_xDisplay, XC_top_left_arrow) ; // not the right arrow...
    }
    XDefineCursor (_xDisplay, _xWindow, cursor) ;
  }

  void
  glWindow_GLX::warpCursor(int x, int y) {
    XWarpPointer(_xDisplay, _xWindow, _xWindow, 0, 0, 0, 0, x, y);
  }

  void
  glWindow_GLX::keyboardAutoRepeat(bool activate) {
    if (activate)
	 XAutoRepeatOn(_xDisplay) ;
    else
	 XAutoRepeatOff(_xDisplay) ;
  }

  void
  glWindow_GLX::raise(void) {
    XRaiseWindow(_xDisplay, _xWindow) ;
  }

  void
  glWindow_GLX::lower(void) {
    XLowerWindow(_xDisplay, _xWindow) ;
  }

  // ------------------------------------------------------------------

  void
  glWindow_GLX::makeCurrent(void) {
    if (!_mapped) map() ;
    glXMakeCurrent(_xDisplay, _xWindow, _glContext) ;
  }

  // ------------------------------------------------------------------

  void
  glWindow_GLX::swapBuffers(void) {
    glXSwapBuffers(_xDisplay, _xWindow) ; // _xWindow was glXGetCurrentDrawable() before
  }

  // ------------------------------------------------------------------

  void
  glWindow_GLX::react(Observable*) {
    // std::cerr << "glWindow_GLX::react " << std::endl ;
    notifyObservers() ;
  }

  bool
  glWindow_GLX::getNextEvent(event *e) {
    static Atom wmDeleteWindow = XInternAtom(_xDisplay, "WM_DELETE_WINDOW", False);
    bool eventHandled = false;

    while(!eventHandled) {
      if (!XPending(_xDisplay)) return false ;
      eventHandled = true;
      XEvent xe ;
      XNextEvent(_xDisplay, &xe) ;
      switch (xe.type) {
      case ConfigureNotify:
	   e->type = glWindow::event::configure ;
	   e->width = xe.xconfigure.width ;
	   e->height = xe.xconfigure.height ;
	   e->x = xe.xconfigure.x ;
	   e->y = xe.xconfigure.y ;
	   e->time = CurrentTime;
	   break ;
      case Expose:
	   e->type = glWindow::event::expose ;
	   e->time = CurrentTime;
	   break ;
      case DestroyNotify:
	   e->type = glWindow::event::destroy;
	   e->time = CurrentTime;
	   break;
      case ClientMessage:
	   if ((unsigned)xe.xclient.data.l[0] == wmDeleteWindow) {
		e->type = glWindow::event::destroy;
	   }
	   e->time = CurrentTime;
	   break;
	   // Ignore Enter/Leave events generated by grabbing or ungrabbing the
	   // pointer if this is needed a new member may be added to e 
      case EnterNotify:
	   if (xe.xcrossing.mode != NotifyNormal)
		eventHandled = false;  // pointer ungrab
	   else
		e->type = glWindow::event::enter ;
	   e->time = xe.xcrossing.time;
	   break ;
      case LeaveNotify:
	   if (xe.xcrossing.mode != NotifyNormal)
		eventHandled = false; // pointer grab
	   else
		e->type = glWindow::event::leave ;
	   break ;
      case ButtonPress:
      case ButtonRelease:
	   e->type = (xe.type==ButtonPress) ? glWindow::event::buttonPress : glWindow::event::buttonRelease ;
	   e->x = xe.xbutton.x ;
	   e->y = xe.xbutton.y ;
	   e->button = xe.xbutton.button ;
	   e->time = xe.xbutton.time;
	   break ;
      case MotionNotify:
	   e->type = glWindow::event::pointerMotion ;
	   while (XCheckTypedWindowEvent(_xDisplay, xe.xmotion.window, MotionNotify, &xe)) ;
	   e->x = xe.xmotion.x ;
	   e->y = xe.xmotion.y ;
	   e->time = xe.xmotion.time;
	   break ;
      case KeyPress:
      case KeyRelease: {
	   KeySym ks ;
	   char keyname[256] ;
	   XLookupString(&xe.xkey, keyname, 256, &ks, NULL) ;
	   if (IsModifierKey(ks)) {
		ks = XKeycodeToKeysym(_xDisplay, xe.xkey.keycode, 0);
		XLookupString(&xe.xkey, keyname, 256, &ks, NULL) ;
	   }
	   e->type = (xe.type==KeyPress) ? glWindow::event::keyPress : glWindow::event::keyRelease ;
	   e->keysym = (unsigned long)ks ;
	   e->keystr = keyname ;
	   e->time = xe.xkey.time;
      } break ;

	   // not handled events selected by StructureNotifyMask
      case CirculateNotify:
      case GravityNotify:
      case MapNotify:
      case ReparentNotify:
      case UnmapNotify:
	   eventHandled = false;
	   break ;

	   // not handled events always selected
      case MappingNotify:
      case SelectionClear:
      case SelectionNotify:
      case SelectionRequest:
	   eventHandled = false;
	   break ;

	   // should not happen
      default:
	   std::cerr << "glWindow_GLX: X event type " << xe.type << std::endl ;
	   eventHandled = false;
      }

      if (debugEvents) {
	   std::cerr << "glWindow_GLX event: " ;
	   e->debug(std::cerr) ;
	   std::cerr << std::endl ;
      }
    } /* while */

    return true ;
  }

  // ------------------------------------------------------------------

}
