/*
 *
 * nucleo/image/sink/qtMovImageSink.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/sink/qtMovImageSink.H>
#include <nucleo/image/encoding/Conversion.H>

// ---------------------------------------------------

extern "C" {
  OSErr NativePathNameToFSSpec(char *, FSSpec *, long) ;
}

namespace nucleo {

  // ---------------------------------------------------

  qtMovImageSink::qtMovImageSink(const URL &url) {
    _filename = url.isOpaque() ? url.getOpaque() : url.getPath() ;
    std::string query = url.getQuery() ;
    std::string codec ;
    URL::getQueryArg(query, "codec", &codec) ;
    _codec = kSorenson3CodecType ;
    if (codec=="dvpal") {
	 _codec = kDVCPALCodecType ;
	 std::cerr << "qtMovImageSink: DV PAL" << std::endl ;
    } else if (codec=="dvntsc") {
	 _codec = kDVCNTSCCodecType ;
	 std::cerr << "qtMovImageSink: DV NTSC" << std::endl ;
    } else if (codec=="mpeg")
	 _codec = kMpegYUV420CodecType ;
    else if (codec=="mpeg4")
	 _codec = kMPEG4VisualCodecType ;
    else if (codec=="indeo4")
	 _codec = kIndeo4CodecType ;
    else if (codec=="sorenson3")
	 _codec = kSorenson3CodecType ;
    else if (codec!="") {
	 std::cerr << "qtMovImageSink: unknown codec '" << codec << "'" << std::endl ;
    }
    std::string framerate ;
    if (URL::getQueryArg(query, "framerate", &framerate))
	 _framerate = atoi(framerate.c_str()) ;
    else
	 _framerate = 25 ;

    _movie = NULL ;
    _track = NULL ;
    _media = NULL ;
    _resRefNum = 0 ;

    _state = BEZERK ;

    EnterMovies() ;
  }

  qtMovImageSink::~qtMovImageSink() {
    stop() ;
  }

  bool
  qtMovImageSink::start(void) {
    FSSpec theFSSpec;
    OSErr err = NativePathNameToFSSpec((char *)_filename.c_str(),
							    &theFSSpec, 0 /* flags */) ;

    err = CreateMovieFile(&theFSSpec,
					 FOUR_CHAR_CODE('TVOD'), // Could use a nucleo FOUR_CHAR_CODE...
					 smCurrentScript,
					 createMovieFileDeleteCurFile|createMovieFileDontCreateResFile,
					 &_resRefNum, &_movie);
    if (err != noErr) {
	 std::cerr << "qtMovImageSink: CreateMovieFile failed (" << err << ")" << std::endl ;
	 if (_resRefNum != 0) CloseMovieFile(_resRefNum);
	 if (_movie != NULL) DisposeMovie(_movie);
	 return false ;
    }

    _state = OPENED ;
    _chrono.start() ;
    return true ;
  }

  bool
  qtMovImageSink::isActive(void) {
    return (_state==OPENED) ;
  }

  bool
  qtMovImageSink::handle(Image *img) {
    if (_state!=OPENED) return false ;

    // CompressImage fails with Image::RGB...
    Image::Encoding qtEncoding = Image::ARGB ;
    OSType pixelFormat = k32ARGBPixelFormat ;

    if (!convertImage(img, &_img, qtEncoding)) return false ;

    if (_track==NULL) {
	 _track = NewMovieTrack(_movie,
					    FixRatio(_img.getWidth(),1),
					    FixRatio(_img.getHeight(),1),
					    kNoVolume);
	 OSErr err = GetMoviesError();
	 if (err != noErr) {
	   std::cerr << "qtMovImageSink: NewMovieTrack failed (" << err << ")" << std::endl ;
	   if (_resRefNum != 0) CloseMovieFile(_resRefNum);
	   if (_movie != NULL) DisposeMovie(_movie);
	   _state = BEZERK ;
	   return false ;
	 }
		
	 _media = NewTrackMedia(_track, VideoMediaType,
					    1000 /* Time scale */,
					    NULL, 0) ;
	 err = GetMoviesError();
	 if (err != noErr) {
	   std::cerr << "qtMovImageSink: NewTrackMedia failed (" << err << ")" << std::endl ;
	   if (_resRefNum != 0) CloseMovieFile(_resRefNum);
	   if (_movie != NULL) DisposeMovie(_movie);
	   return false ;
	 }
    
	 err = BeginMediaEdits(_media);
	 if (err != noErr) {
	   std::cerr << "qtMovImageSink: BeginMediaEdits failed (" << err << ")" << std::endl ;
	   if (_resRefNum != 0) CloseMovieFile(_resRefNum);
	   if (_movie != NULL) DisposeMovie(_movie);
	   return false ;
	 }
    }

    // --------------------------------

    Rect bounds ;
    bounds.top = bounds.left = 0 ;
    bounds.right = _img.getWidth() ;
    bounds.bottom = _img.getHeight() ;

    GWorldPtr gworld ;
    OSErr err = QTNewGWorldFromPtr(&gworld, pixelFormat, &bounds,
							NULL, NULL, 0,
							_img.getData(), 4*_img.getWidth()) ;

    if (err != noErr) {
	 std::cerr << "qtMovImageSink: QTNewGWorldFromPtr failed (" << err << ")" << std::endl ;
	 return false ;
    }

    PixMapHandle pixmap = GetGWorldPixMap(gworld) ;
    if (pixmap == NULL) {
	 std::cerr << "qtMovImageSink: GetGWorldPixMap failed (" << err << ")" << std::endl ;
	 return false ;
    }

    long maxComprSize = 0L ;
    err = GetMaxCompressionSize(pixmap,
						  &bounds, 
						  0, // let ICM choose depth
						  codecNormalQuality, 
						  _codec, 
						  (CompressorComponent)anyCodec,
						  &maxComprSize);
    if (err != noErr) {
	 std::cerr << "qtMovImageSink: GetMaxCompressionSize failed (" << err << ")" << std::endl ;
	 return false ;
    } 

    Handle myComprDataHdl = NewHandle(maxComprSize);
    if (myComprDataHdl == NULL) {
	 std::cerr << "qtMovImageSink: NewHandle failed (" << err << ")" << std::endl ;
	 return false ;
    }
    HLockHi(myComprDataHdl);

    Ptr myComprDataPtr = *myComprDataHdl;

    ImageDescriptionHandle myImageDesc = (ImageDescriptionHandle)NewHandle(4);
    if (myImageDesc == NULL) {
	 std::cerr << "qtMovImageSink: ImageDescriptionHandle NewHandle failed (" << err << ")" << std::endl ;
	 return false ;
    }

    err = CompressImage(pixmap, 
				    &bounds, 
				    codecNormalQuality,
				    _codec,
				    myImageDesc, 
				    myComprDataPtr);
    if (err != noErr) {
	 std::cerr << "qtMovImageSink: CompressImage failed (" << err << ", " << GetMoviesError() << ")" << std::endl ;
	 return false ;
    }

    err = AddMediaSample(_media, 
					myComprDataHdl,
					0, // no offset in data
					(**myImageDesc).dataSize, 
					1000/_framerate, // frame duration
					(SampleDescriptionHandle)myImageDesc, 
					1,	 // one sample
					0,	 // self-contained samples
					NULL);
    if (err != noErr) {
	 std::cerr << "qtMovImageSink: AddMediaSample failed (" << err << ")" << std::endl ;
	 return false ;
    }

    if (myImageDesc != NULL) DisposeHandle((Handle)myImageDesc) ;
    if (myComprDataHdl != NULL) DisposeHandle(myComprDataHdl) ;
    if (gworld != NULL) DisposeGWorld(gworld) ;

    _chrono.tick() ;
    return true ;
  }

  bool
  qtMovImageSink::stop(void) {
    _state = CLOSED ;

    OSErr err = EndMediaEdits(_media);
    if (err != noErr) {
	 std::cerr << "qtMovImageSink: EndMediaEdits failed (" << err << ")" << std::endl ;
	 if (_resRefNum != 0) CloseMovieFile(_resRefNum);
	 if (_movie != NULL) DisposeMovie(_movie);
	 return false ;
    }
    
    err = InsertMediaIntoTrack(_track, 0, 0, GetMediaDuration(_media), fixed1);
    if (err != noErr) {
	 std::cerr << "qtMovImageSink: InsertMediaIntoTrack failed (" << err << ")" << std::endl ;
	 if (_resRefNum != 0) CloseMovieFile(_resRefNum);
	 if (_movie != NULL) DisposeMovie(_movie);
	 return false ;
    }

    short myResID = movieInDataForkResID;
    err = AddMovieResource(_movie, _resRefNum, &myResID, NULL);
    if (err != noErr) {
	 std::cerr << "movieInDataForkResID: AddMovieResource failed (" << err << ")" << std::endl ;
	 if (_resRefNum != 0) CloseMovieFile(_resRefNum);
	 if (_movie != NULL) DisposeMovie(_movie);
	 return false ;
    }

    _chrono.stop() ;
    return true ;
  }

  // ---------------------------------------------------

}

