/*
 *
 * nucleo/image/source/ImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/source/ImageSource.H>

#ifdef __APPLE__
#include <nucleo/image/source/qtMovImageSource.H>
#include <nucleo/image/source/qtSgImageSource.H>
#endif

#ifdef __linux__
#include <nucleo/image/source/v4lImageSource.H>
#endif

#include <nucleo/image/source/nudpImageSource.H>
#include <nucleo/image/source/nudppImageSource.H>
#include <nucleo/image/source/nudpcImageSource.H>

#include <nucleo/image/source/serverpushImageSource.H>

#include <nucleo/image/source/imagefileImageSource.H>
#include <nucleo/image/source/nspImageSource.H>

#include <nucleo/image/source/noiseImageSource.H>

#include <nucleo/core/PluginManager.H>

#include <string>
#include <stdexcept>

namespace nucleo {

  // ------------------------------------------------------------------

  bool
  ImageSource::getLatestImage(Image *img) {
    return getNextImage(img) ;
  }

  bool
  ImageSource::waitForImage(Image *img, bool latest) {
    while (isActive()) {
	 bool haveImage = latest ? getLatestImage(img) : getNextImage(img) ;
	 if (haveImage) return true ;
	 ReactiveEngine::step() ;
    }
    return false ;
  }

  // ------------------------------------------------------------------

  bool
  ImageSource::getImage(const char *uri, Image *img, Image::Encoding enc) {
    ImageSource *source = ImageSource::create(uri, enc) ;
    source->start() ;
    bool gotTheImage = source->waitForImage(img) ;
    if (gotTheImage && img->dataIsLinked()) img->acquireData() ;
    delete source ;
    return gotTheImage ;
  }

  // ------------------------------------------------------------------

  ImageSource *
  ImageSource::create(const char *u, Image::Encoding e) {
    if (!u) throw std::runtime_error("Can't create an ImageSource from an empty URL...") ;

    URL url(u) ;
    std::string scheme = url.getScheme() ;

    // -------------------------------------------------------

    std::string tag ;

#ifdef __APPLE__
    if (scheme == "videoin")
	 return new qtSgImageSource(e, url) ;
#endif
#ifdef __linux__
    if (scheme == "videoin")
	 return new v4lImageSource(e, url) ;
#endif
    if (scheme == "nudp")
      return new nudpImageSource(e, url) ;
    else if ((scheme == "nudp+") || (scheme == "nudpp"))
	 return new nudppImageSource(e, url) ;
    else if (scheme == "nudpc")
	 return new nudpcImageSource(e, url) ;
    else if (scheme=="http")
	 return new serverpushImageSource(e, url) ;
    else if (scheme=="noise")
	 return new noiseImageSource(e, url) ;
    else if ((scheme=="file") || (scheme=="")) {
	 std::string filename = url.isOpaque() ? url.getOpaque() : url.getPath() ;
	 const char *extension = getExtension(filename.c_str()) ;
	 if (!extension) throw std::runtime_error("createImageSource: unknown file type '"+filename+"'") ;
	 cistring iext = extension ;
	 if ((iext == ".vss") || (iext == ".nsp")) return new nspImageSource(e, url) ;
	 if (iext == ".jpg") return new imagefileImageSource(e, url, Image::JPEG) ;
	 if (iext == ".jpeg") return new imagefileImageSource(e, url, Image::JPEG) ;
	 if (iext == ".png") return new imagefileImageSource(e, url, Image::PNG) ;
	 if (iext == ".pam") return new imagefileImageSource(e, url, Image::PAM) ;
#ifdef __APPLE__
	 if (iext == ".mov") return new qtMovImageSource(e, url) ;
	 if (iext == ".dv") return new qtMovImageSource(e, url) ;
	 if (iext == ".mpg") return new qtMovImageSource(e, url) ;
	 if (iext == ".mp4") return new qtMovImageSource(e, url) ;
	 if (iext == ".avi") return new qtMovImageSource(e, url) ;
	 if (iext == ".swf") return new qtMovImageSource(e, url) ;
#endif
	 tag = std::string("extension=")+extension ;
    } else {
	 tag = std::string("scheme=")+scheme ;
    }

    ImageSourceFactory factory = (ImageSourceFactory) PluginManager::getSymbol("ImageSource::create",tag) ;
    return (ImageSource*) ((*factory)(&url, e)) ;
  }

  // ------------------------------------------------------------------

}
