/*
 *
 * nucleo/image/source/qtSgImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/source/qtSgImageSource.H>
#include <nucleo/image/encoding/Conversion.H>

#define CERR_DEBUG 0

namespace nucleo {

  // --------------------------------------------------------------------------

  pascal OSErr
  sgDataProc(SGChannel c,
		   Ptr p, long len, long *offset,
		   long chRefCon, TimeValue time,
		   short writeType, long refCon) { 
    qtSgImageSource *source = (qtSgImageSource*)refCon ;
    if (!source || !source->_pGWorld) return -1 ;

    if (source->_haveImage) {
#if CERR_DEBUG
	 std::cerr << "-" << std::flush ;
#endif
	 return -1 ;
    }
#if CERR_DEBUG
    std::cerr << "+" << std::flush ;
#endif

    ComponentResult err = noErr ;

    if (source->_decomSeq == 0) {
	 ImageDescriptionHandle imageDesc = (ImageDescriptionHandle)NewHandle(0) ;
	 err = SGGetChannelSampleDescription(c,(Handle)imageDesc) ;
	 if (err!=noErr) {
	   std::cerr << "SGGetChannelSampleDescription failed (" << err << ")" << std::endl ;
	   return err ;
	 }

	 err = DecompressSequenceBeginS(&(source->_decomSeq),
							  imageDesc,
							  p, len,
							  source->_pGWorld, nil,
							  &(source->_bounds),
							  nil,	// transformation matrix
							  srcCopy,	// transfer mode specifier
							  nil,	// clipping region
							  codecFlagDontUseNewImageBuffer,		// flags
							  codecNormalQuality,
							  bestSpeedCodec) ;
	 if (err!=noErr) {
	   std::cerr << "DecompressSequenceBeginS failed (" << err << ")" << std::endl ;
	   return err ;
	 }
            
	 DisposeHandle((Handle)imageDesc) ;
	 imageDesc = nil ;
    } 

    err = DecompressSequenceFrameWhen(source->_decomSeq,
							   p, len,
							   0, 0,
							   nil, nil) ;
    if (err!=noErr) {
	 std::cerr << "DecompressSequenceFrameWhen failed (" << err << ")" << std::endl ;
	 return err ;
    }

    source->_haveImage=true;
    source->notifyObservers() ;

    return noErr ;
  }

  // --------------------------------------------------------------------------

  qtSgImageSource::qtSgImageSource(Image::Encoding e, const URL &url) {
    _sg = nil ;
    _pGWorld = nil ;
    _bounds.top = _bounds.left = _bounds.right = _bounds.bottom = 0 ;
    _decomSeq = nil ;

    int qtEncoding ;

    // QTNewGWorldFromPtr fails with k32ABGRPixelFormat and k32RGBAPixelFormat
    // DecompressSequenceFrameWhen fails with kYUV420PixelFormat

    if (e==Image::RGB) {
	 _nucleo_encoding = e ;
	 _qt_encoding = Image::RGB ;
	 qtEncoding = k24RGBPixelFormat ;
    } else {
	 _nucleo_encoding = e==Image::PREFERRED ? Image::ARGB : e ; 
	 _qt_encoding = Image::ARGB ;
	 qtEncoding = k32ARGBPixelFormat ;
    }

    // std::cerr << Image::getEncodingName(_qt_encoding) << " --> " << Image::getEncodingName(_nucleo_encoding) << std::endl ;

    _haveImage = false ;

    _state = BEZERK ;

    // ------------------------------------------

    std::string query = url.getQuery() ;

    std::string size ;
    URL::getQueryArg(query, "size", &size) ;
    bool showDialog = URL::getQueryArg(query, "dialog") ;

    std::string arg ;

    double framerate = 0 ;
    if (URL::getQueryArg(query, "framerate", &arg)) framerate = atof(arg.c_str()) ;
    if (URL::getQueryArg(query, "pause", &arg)) framerate = 1.0/atof(arg.c_str()) ;

    if (!URL::getQueryArg(query, "quality", &_quality))
	 _quality = 100 ;

    // ------------------------------------------

    EnterMovies();

#if CERR_DEBUG
    std::cerr << "Opening sequence grabber" << std::endl ;
#endif

    _sg = OpenDefaultComponent(SeqGrabComponentType, 0);

    OSErr err = SGInitialize(_sg) ;
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: SGInitialize failed (" << err << ")" << std::endl ;
	 return ;
    }

    err = SGSetDataRef(_sg, 0, 0, seqGrabDontMakeMovie);
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: SGSetDataRef failed (" << err << ")" << std::endl ;
	 return ;
    }

    err = SGSetDataProc(_sg,NewSGDataUPP(sgDataProc),(long)this);
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: SGSetDataProc failed (" << err << ")" << std::endl ;
	 return ;
    }

    // ------------------------------------------

#if CERR_DEBUG
    std::cerr << "Opening channel" << std::endl ;
#endif

    SGChannel channel = nil ;
    err = SGNewChannel(_sg, VideoMediaType, &channel);
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: SGNewChannel failed (" << err << ")" << std::endl ;
	 return ;
    }

    err = SGSetChannelUsage(channel, seqGrabRecord|seqGrabLowLatencyCapture) ;
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: SGSetChannelUsage failed (" << err << ")" << std::endl ;
	 return ;
    }

    if (showDialog) {
	 SGSettingsDialog(_sg, channel, 0, nil, 0, nil, nil);
    } else {
	 SGSetVideoCompressorType(channel, kPlanarRGBCodecType) ;
	 Fixed fr = (Fixed)(framerate*65536) ;
	 err = SGSetFrameRate(channel, fr) ;
	 if (err!=noErr) {
	   if (err) std::cerr << "qtSgImageSource: SGSetFrameRate failed (" << err << ")" << std::endl ;
	   return ;
	 }
    }
#if 0
    Fixed fr = 0 ; err = SGGetFrameRate(channel, &fr) ;
    std::cerr << "Framerate is " << fr << " (" << fr/65536 << " fps)" << std::endl ;
#endif

    // ------------------------------------------

    err = SGGetSrcVideoBounds(channel,&_bounds) ;
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: SGGetSrcVideoBounds failed (" << err << ")" << std::endl ;
	 return ;
    }

    if (size=="") size = "SIF" ;
    int iwin = 0 ;
    for (iwin=0; Image::StandardSizes[iwin].name!=0; ++iwin)
	 if (size==Image::StandardSizes[iwin].name) break ;
    if (Image::StandardSizes[iwin].name==0) {
	 std::cerr << "qtSgImageSource> Image size '" << size << "' unknown..." << std::endl ;
	 return ;
    }
    _bounds.right = Image::StandardSizes[iwin].width ;
    _bounds.bottom = Image::StandardSizes[iwin].height ;

    err = SGSetChannelBounds(channel, &_bounds);
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: SGSetChannelBounds failed (" << err << ")" << std::endl ;
	 return ;
    }

#if CERR_DEBUG
    std::cerr << "Creating GWorld" << std::endl ;
#endif

    unsigned int width = _bounds.right - _bounds.left;
    unsigned int height = _bounds.bottom - _bounds.top;
    long rowBytes = (long)(width*_img.getBytesPerPixel(_qt_encoding)) ;

    _img.prepareFor(width, height, _qt_encoding) ;
    err = QTNewGWorldFromPtr(&_pGWorld, qtEncoding, &_bounds, NULL, NULL, 0,
					    _img.getData(), rowBytes) ;
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: QTNewGWorldFromPtr failed (" << err << ")" << std::endl ;
	 return ;
    }

    err = SGSetGWorld(_sg,(CGrafPtr)_pGWorld,GetMainDevice());
    if (err!=noErr) {
	 if (err) std::cerr << "qtSgImageSource: SGSetGWorld failed (" << err << ")" << std::endl ;
	 return ;
    }

    _state = READY ;
  }

  qtSgImageSource::~qtSgImageSource() {
    stop();
    CloseComponent(_sg);
    if (_decomSeq) CDSequenceEnd(_decomSeq);
    if (_pGWorld) {
	 DisposeGWorld(_pGWorld);
	 _pGWorld = nil;
    }
  }

  bool
  qtSgImageSource::start() {  
    if (_state!=READY) return false ;

#if CERR_DEBUG
    std::cerr << "SGStartRecord" << std::endl ;
#endif
    if (SGStartRecord(_sg)==noErr) {
	 _state = ACTIVE ;
	 _chrono.start() ;
	 _tk = TimeKeeper::create(0, true) ;
	 subscribeTo(_tk) ;
	 return true ;
    }

    _state = BEZERK ;
    return false ;
  }

  void
  qtSgImageSource::react(Observable* obs) {
    if (!_state==ACTIVE) return ;
    SGIdle(_sg) ;
  }

  bool
  qtSgImageSource::getNextImage(Image * img) {
    if (_state!=ACTIVE || !_haveImage) return false ;
    convertImage(&_img, img, _nucleo_encoding, _quality) ;
    _chrono.tick() ;
    _haveImage = false ;
    return true ;
  }

  bool
  qtSgImageSource::stop() {
    if(_state==ACTIVE) {
	 SGStop(_sg);
	 _state = BEZERK ;
	 unsubscribeFrom(_tk) ;
	 delete _tk ;
	 return true ;
    }
    return false ;
  }

  // --------------------------------------------------------------------------

}
