/*
 *
 * nucleo/image/source/serverpushImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/source/serverpushImageSource.H>

#include <nucleo/utils/FileUtils.H>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdexcept>

namespace nucleo {

  // ------------------------------------------------------------------

  serverpushImageSource::serverpushImageSource(Image::Encoding encoding, const URL &url) : ImageSource() {
    _encoding = encoding ; 

    std::string query=url.getQuery() ;
    if (!URL::getQueryArg(query, "quality", &_quality))
	 _quality = 100 ;

    std::string tmp ;
    _hostname = url.getHost() ;
    if (_hostname=="") _hostname="localhost" ;
    _port = atoi(url.getPort().c_str()) ;
    if (!_port) _port = 80 ;

    _request = "GET " ;

    tmp = url.getPath() ;
    if (tmp!="") _request.append(tmp) ; else _request.append("/") ;

    tmp = url.getQuery() ;
    if (tmp!="") {
	 _request.append("?") ;
	 _request.append(tmp) ;
    }

    _request.append(" HTTP/1.1") ;
    _request.append(oneCRLF) ;
#if 0
     _request.append("Connection: keep-alive") ; _request.append(oneCRLF) ;
#else
    _request.append("Connection: close") ; _request.append(oneCRLF) ;
#endif
    _request.append("User-Agent: Mozilla/1.1") ; _request.append(oneCRLF) ;
    _request.append("Accept: image/jpeg") ; _request.append(oneCRLF) ;
    _request.append("Host: ") ; _request.append(url.getHost()) ; _request.append(oneCRLF) ;
    _request.append(oneCRLF) ;

    _state = CLOSED ;
    _connection = 0 ;
  }

  bool
  serverpushImageSource::start(void) {
    if (_state!=CLOSED) return false ;

    _msg.reset() ;
    try {
	 _connection = new TcpConnection(_hostname.c_str(), _port) ;
    } catch (std::runtime_error e) {
	 std::cerr << "serverpushImageSource: " << e.what() << std::endl ;
	 return false ;
    } catch (...) {
	 return false ;
    }

    subscribeTo(_connection) ;
	
    try {
	 _connection->send((const char*) _request.c_str(), _request.length()) ;
    } catch (std::runtime_error e) {
	 std::cerr << "serverpushImageSource: " << e.what() << std::endl ;
	 delete _connection ;
	 return false;
    } catch (...) {
	 return false ;
    }
	
    _state = OPEN ;
    _chrono.start() ;
    return true ;
  }

  void
  serverpushImageSource::react(Observable *obs) {
    switch(_state) {
    case OPEN:
	 if (obs==_connection) {
	   int fd = _connection->getFd() ;
	   if (_msg.feedFromStream(fd, getavail(fd))<1) _state = BUFFERED ;
	   notifyObservers() ;
	 }
	 break ;
    case BUFFERED:
	 if (!_msg.dataWaiting()) stop() ;
	 notifyObservers() ;
	 break ;
    default:
	 break ;
    }

  }

  bool
  serverpushImageSource::getLatestImage(Image *img) {
    if (_state==CLOSED) return false ; 

    bool loop, foundOne=false ;
    do {
	 HttpMessage::State s = _msg.parseData() ;
	 loop = false ;

	 if (s==HttpMessage::NEED_BODY && _state==BUFFERED) s=_msg.completeData() ;

	 if (s==HttpMessage::COMPLETE) {
	   std::string smimetype, swidth, sheight ;
	   unsigned int width=0, height=0 ;
	   Image::Encoding encoding = Image::JPEG ;

	   if (_msg.getHeader("nucleo-image-width", swidth)) width = atoi(swidth.c_str()) ;
	   if (_msg.getHeader("nucleo-image-height", sheight)) height = atoi(sheight.c_str()) ;
	   if (_msg.getHeader("content-type", smimetype)) encoding = Image::getEncodingByMimeType(smimetype) ;

	   // std::cerr << smimetype << " " << swidth << "x" << sheight << std::endl ;

	   std::string const &body = _msg.body() ;
	   img->setEncoding(encoding) ;
	   img->setDims(width, height) ;
	   img->setData((unsigned char *)body.c_str(), body.length(), Image::NONE) ;
	   convertImage(img, _encoding) ;

	   // If no conversion was necessary, we must copy the data since it
	   // will be overwritten by the _msg.next() call below
	   if (img->getData()==(unsigned char *)body.c_str())
		img->acquireData() ;

	   _chrono.tick() ;
	   _msg.next() ;   
	   loop = foundOne = true ;
	 }

    } while (loop) ;

    if (!foundOne && _state==BUFFERED) _state = CLOSED ;

    return foundOne ;
  }

  bool
  serverpushImageSource::getNextImage(Image *img) {
    if (_state==CLOSED) return false ; 

    HttpMessage::State s = _msg.parseData() ;
    if (s==HttpMessage::NEED_BODY && _state==BUFFERED) s=_msg.completeData() ;

    if (s!=HttpMessage::COMPLETE) {
	 if (_state==BUFFERED) _state = CLOSED ;
	 return false ;
    }

    std::string const &body = _msg.body() ;
    img->setEncoding(Image::JPEG) ;
    img->setDims(0,0) ;
    img->setData((unsigned char *)body.c_str(), body.length(), Image::NONE) ;
    convertImage(img, _encoding) ;

    // If no conversion was necessary, we must copy the data since it
    // will be overwritten by the _msg.next() call below
    if (img->getData()==(unsigned char *)body.c_str())
	 img->acquireData() ;

    _chrono.tick() ;
    _msg.next() ;	 
    return true ;
  }

  bool
  serverpushImageSource::stop(void) {
    if (_state!=CLOSED) {
	 _chrono.stop() ;
	 unsubscribeFrom(_connection) ;
	 delete _connection ;
	 _state = CLOSED ;
	 return true ;
    }
    return false ;
  }

  // ------------------------------------------------------------------

}
