/*
 *
 * nucleo/network/tcp/TcpConnection.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/network/tcp/TcpConnection.H>
#include <nucleo/network/tcp/TcpUtils.H>
#include <nucleo/utils/FileUtils.H>
#include <nucleo/core/FileKeeper.H>
#include <nucleo/core/TimeKeeper.H>
#include <nucleo/core/ReactiveEngine.H>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <sys/errno.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

namespace nucleo {

  // -------------------------------------------------------------------------

  TcpConnection::TcpConnection(int socket, bool close) {
    if (socket==-1)
	 throw std::runtime_error("TcpConnection: bad socket (-1)") ;
    _socket = socket ;
    _closeSocketOnDestroy = close ;
    setDefaultTcpSocketOptions(_socket, false) ; 

    _tcpw = FileKeeper::create(_socket, FileKeeper::R) ;
    subscribeTo(_tcpw) ;
  }    

  TcpConnection::TcpConnection(const char* hostname, int port, bool close) {
    _socket = -1 ;
    _closeSocketOnDestroy = false ;

    struct sockaddr_in addr ;
    addr.sin_family = AF_INET ;
    addr.sin_port = htons(port) ;
    memset(addr.sin_zero, 0, 8) ;

    if (inet_aton(hostname, &(addr.sin_addr))==0) {
	 struct hostent *h = gethostbyname(hostname) ;
	 if (!h)
	   throw std::runtime_error(std::string("TcpConnection: gethostbyname failed (")+hostname+")") ;
#if 0
	 std::cerr << "h_name: " << h->h_name << std::endl ;
	 std::cerr << "h_length: " << h->h_length << std::endl ;
	 for (int i=0; i<h->h_length; ++i) {
	   unsigned char *ptr = (unsigned char *)h->h_addr_list[i] ;
	   std::cerr << "h_addr_list[" << i << "]: " << (int)ptr[0] << "." << (int)ptr[1] << "." << (int)ptr[2] << "." << (int)ptr[3] << std::endl ;
	 }
#endif
	 addr.sin_addr.s_addr = *(long*)h->h_addr ;
    }

    _socket = socket(AF_INET, SOCK_STREAM, 0) ;

    if( connect(_socket, (sockaddr *)&addr, sizeof(addr)) ) {
	 std::stringstream msg ;
	 msg << "TcpConnection: connect failed (" << hostname << ":" << port << ")" ;
	 throw std::runtime_error(msg.str()) ;
    }

    setDefaultTcpSocketOptions(_socket, false) ; 
    _closeSocketOnDestroy = close ;

    _tcpw = FileKeeper::create(_socket, FileKeeper::R) ;
    subscribeTo(_tcpw) ;
  }

  // -------------------------------------------------------------------------

  unsigned int
  TcpConnection::send(const char *data, unsigned int length, bool complete) {
    if (!length) return 0 ;

    int n ;
    unsigned int bytesToWrite = length ;
    do {
	 n = write(_socket, data, bytesToWrite) ;
	 if (n==-1) throw std::runtime_error("TcpConnection: write failed") ;
	 bytesToWrite -= n ;
    } while (n && bytesToWrite && complete) ;

    return length-bytesToWrite ;
  }

  // -------------------------------------------------------------------------

  void
  TcpConnection::react(Observable *obs) {
    if (obs==_tcpw) notifyObservers() ;
  }

  // -------------------------------------------------------------------------

  unsigned int
  TcpConnection::receive(char *data, unsigned int length, bool complete) {
    if (!length) return 0 ;

    int n ;
    unsigned int bytesToRead = length ;
    do {
	 n = read(_socket, (void *)data, bytesToRead) ;
	 if (n==-1) throw std::runtime_error("TcpConnection: read failed") ;
	 bytesToRead -= n ;
    } while (n && bytesToRead && complete) ;

    return length-bytesToRead ;
  }

  // -------------------------------------------------------------------------

  std::string
  TcpConnection::machineLookUp(void) {
    sockaddr_in addr;
    struct hostent *h ;
    socklen_t lenaddr = sizeof(sockaddr_in);

    if (getpeername(_socket,(struct sockaddr *)&addr,&lenaddr)==-1)
	 throw std::runtime_error("TcpConnection: machineLookUp failed") ;

    if ((h=gethostbyaddr((char *)&addr.sin_addr, sizeof(addr.sin_addr), AF_INET)))
	 return std::string(h->h_name) ;
    else {
	 unsigned char *v = (unsigned char *)&addr.sin_addr.s_addr ;
	 char result[256] ;
	 sprintf(result, "%d.%d.%d.%d",v[0],v[1],v[2],v[3]) ;
	 return std::string(result) ;
    }
  }

  // -------------------------------------------------------------------------

  std::string
  TcpConnection::userLookUp(int millisecs) {
    sockaddr_in myaddr;
    socklen_t lenmyaddr = sizeof(sockaddr_in);
    sockaddr_in hisaddr;
    socklen_t lenhisaddr = sizeof(sockaddr_in);

    if (getsockname(_socket, (sockaddr *)&myaddr, &lenmyaddr)==-1) return "?" ;
    if (getpeername(_socket,(sockaddr *)&hisaddr,&lenhisaddr)==-1) return "?" ;

    struct hostent *h = gethostbyaddr((char *)&hisaddr.sin_addr, sizeof(hisaddr.sin_addr), AF_INET) ;
    if (h==0) return "?" ;

    int isock = socket(AF_INET, SOCK_STREAM, 0) ;
    struct sockaddr_in addr ;
    addr.sin_family = AF_INET ;
    addr.sin_port = htons(113) ;
    addr.sin_addr.s_addr = *(long*)h->h_addr ;
    memset(addr.sin_zero, 0, 8) ;

    setblocking(isock, 0) ;

    if( connect(isock, (sockaddr *)&addr, sizeof(addr)) ) {
	 if (errno!=EINPROGRESS) return "?" ;

	 // The socket is non-blocking and the connection cannot be
	 // completed immediately.  It is possible to select for
	 // completion by selecting the socket for writing.

	 FileKeeper *fk = FileKeeper::create(isock, FileKeeper::W) ;
	 TimeKeeper *tk = TimeKeeper::create(millisecs) ;
	 int connected ;
	 do {
	   ReactiveEngine::step() ;
	   connected = (fk->getState()&FileKeeper::W) ;
	 } while ( (!connected) && (tk->getState()!=TimeKeeper::TRIGGERED) ) ;
	 delete tk ;
	 delete fk ;
	 if (! connected) return "?" ;
	 
	 int val=0 ;
	 socklen_t len=sizeof(val) ;
	 getsockopt(isock,SOL_SOCKET,SO_ERROR,&val,&len) ;
	 if (val==ETIMEDOUT || val==ECONNREFUSED || val==EHOSTDOWN || val==EHOSTUNREACH)
	   return "?" ;
    }

    char buffer[512] ;
    sprintf(buffer, "%d, %d\n", ntohs(hisaddr.sin_port), ntohs(myaddr.sin_port)) ;
    setblocking(isock, 0) ;
    write(isock, buffer, strlen(buffer)) ;
    setblocking(isock, 1) ;

    int bytesread = read(isock, buffer, 512) ;
    if (!bytesread)  return "?" ;

    while (bytesread>0 && isspace((int)buffer[bytesread-1])) bytesread-- ;
    buffer[bytesread] = '\0';

    int i ;
    for (i=bytesread; i>0 && buffer[i-1]!=':'; i--) ;
    return std::string(buffer+i) ;
  }

  // -------------------------------------------------------------------------

  TcpConnection::~TcpConnection(void) {
    unsubscribeFrom(_tcpw) ;
    delete _tcpw ;
    if (_closeSocketOnDestroy) {
	 shutdown(_socket,2) ;
	 ::close(_socket) ;
    }
  }

  // -------------------------------------------------------------------------

}
