/*
 *
 * mosplugins/ffmpeg/ffmpegImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include "ffmpegImageSource.H"

#include <nucleo/utils/FileUtils.H>
#include <nucleo/image/sink/ImageSink.H>
#include <nucleo/image/encoding/Conversion.H>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdexcept>

// ------------------------------------------------------------------------------------------------------

extern "C" {

  void *ffmpegImageSource_factory(const URL *url, Image::Encoding e) {
    std::string filename = url->isOpaque() ? url->getOpaque() : url->getPath() ;
    const char *extension = getExtension(filename.c_str()) ;
    if (!strcmp(extension,".mpg"))  return new ffmpegImageSource(e, *url,"mpeg1video") ;
    if (!strcmp(extension,".mpeg")) return new ffmpegImageSource(e, *url,"mpeg1video") ;
    throw std::runtime_error("ffmpegImageSource: unknown extension") ;
  }

}

// ------------------------------------------------------------------------------------------------------

static const int BufferSize = 1024*1024 ;

ffmpegImageSource::ffmpegImageSource(Image::Encoding e, const URL &url, const char *format) {
  _encoding = e==Image::PREFERRED ? Image::YpCbCr420 : e ;
  _filename = url.isOpaque() ? url.getOpaque() : url.getPath() ;
  _format = format ;

  std::string query = url.getQuery() ;
  std::string arg = URL::getQueryArg(query, "length") ;
  _length = (arg!="") ? atoi(arg.c_str()) : -1 ;
  arg = URL::getQueryArg(query, "framerate") ;
  _pace = (arg!="") ? (unsigned long)(TimeKeeper::second/atof(arg.c_str())) : 0 ;
  _buffer = new uint8_t [BufferSize] ;
  _bufbegin = BufferSize ;
  _bufend = BufferSize ;

  _state = CLOSED ;
}


bool
ffmpegImageSource::start(void) {
  if (_state!=CLOSED) return false ;

  if (!first_iformat) {
    av_register_all() ;
    avcodec_register_all() ;
  }

  enum CodecID id = CODEC_ID_MPEG1VIDEO ;
  // if (_format=="...") id = ... ;
  _codec = avcodec_find_decoder(id);
  if (!_codec) {
    std::cerr << "ffmpegImageSource: codec not found" << std::endl ;
    return false ;
  }

  // FIXME: we should free this context somewhere
  _codec_context = avcodec_alloc_context() ;

  // we dont send complete frames
  if(_codec->capabilities&CODEC_CAP_TRUNCATED) _codec_context->flags|= CODEC_FLAG_TRUNCATED ;
  
  if (avcodec_open(_codec_context, _codec) < 0) {
    std::cerr << "ffmpegImageSource: could not open codec" << std::endl ;
    return false ; 
  }

  struct stat statinfo ;
  if (stat(_filename.c_str(), &statinfo)!=-1) {
    _fd = open(_filename.c_str(), O_RDONLY) ;
    _state = WAITING ;
    _tk = TimeKeeper::create(_pace, true) ;
    subscribeTo(_tk) ;
    _chrono.start() ;
    return true ;
  }

  std::cerr << "ffmpegImageSource: no such file (" << _filename << ")" << std::endl ;
  close(_fd) ;
  _fd = -1 ;
  return false ;
}

void
ffmpegImageSource::react(Observable *obs) {
  if (obs==_tk) {
    if (_state==WAITING) {
	 for (bool loop=true; loop;) {
	   if (_bufbegin==_bufend) {
		int bytesread = read(_fd, _buffer, BufferSize) ;
		if (bytesread<=0) {
		  stop() ;
		  return ;
		}
		_bufbegin = 0 ;
		_bufend = bytesread ;
	   }

	   int ok = 0 ;
	   int nb = avcodec_decode_video(_codec_context, &_frame, &ok, _buffer+_bufbegin, _bufend-_bufbegin) ;
	   if (nb<0) {
		std::cerr << "ffmpegImageSource: error while decoding frame" << std::endl ;
		stop() ;
		return ;
	   }
	   _bufbegin += nb ;
	   loop = !ok ;
	 }

	 _state = READY ;
	 notifyObservers() ;
    }
  }
}

bool
ffmpegImageSource::getNextImage(Image *img) {
  if (_state!=READY) return false ; 

  unsigned int w = _codec_context->width ;
  unsigned int h = _codec_context->height ;
  unsigned char *data = 0 ;
  unsigned int datasize = 0 ;

  switch (_encoding) {
  case Image::L:
    img->setEncoding(Image::L) ;
    datasize = w*h ;
    data = (unsigned char *)Image::AllocMem(datasize) ;
    for(unsigned int i=0; i<h; ++i)
	 memcpy(data + i*w, _frame.data[0] + i*_frame.linesize[0], w) ;
    break ;
  case Image::YpCbCr420: 
  default: {
	 img->setEncoding(Image::YpCbCr420) ;
	 datasize = (unsigned int)(w*h*1.5) ;
	 data = (unsigned char *)Image::AllocMem(datasize) ;
	 unsigned char *yp=data, *cb=yp+w*h, *cr=cb+(int)(0.25*w*h) ;
	 unsigned int hw=w/2, hh=h/2 ;
	 for (unsigned int line=0; line<h; ++line)
	   memcpy(yp + line*w, _frame.data[0] + line*_frame.linesize[0], w) ;	 
	 for (unsigned int line=0; line<hh; ++line) {
	   memcpy(cb + line*hw, _frame.data[1] + line*_frame.linesize[1], hw) ;	 
	   memcpy(cr + line*hw, _frame.data[2] + line*_frame.linesize[2], hw) ;	 
	 }
  } break ;
  }

  img->setDims(w, h) ;
  img->setData(data, datasize, Image::FREEMEM) ;
  convertImage(img, _encoding) ;

  _chrono.tick() ;

  if (_length>0 && (int)(_chrono.count())==_length) {
    stop() ;
  } else {
    _state = WAITING ;
  }

  return true ;
}

bool
ffmpegImageSource::stop(void) {
  if (_state==CLOSED) return false ;

  _chrono.stop() ;
  avcodec_close(_codec_context) ;
  av_freep(&_codec_context) ;
  close(_fd) ;
  _fd = -1 ;
  _state = CLOSED ;

  unsubscribeFrom(_tk) ;
  delete _tk ;

  notifyObservers() ;
  return true ;
}
